#! /usr/bin/env python


# OpenFisca
# Retreives data from erf foyer
# Creates sif and foyer_aggr


import gc
import logging

import numpy
from pandas import concat, DataFrame, Series
import re

from openfisca_survey_manager.temporary import temporary_store_decorator

from openfisca_france_data.erfs.input_data_builder.base import (
    year_specific_by_generic_data_frame_name)
from openfisca_france_data.utils import (
    print_id, build_cerfa_fields_by_column_name)
from openfisca_survey_manager.survey_collections import SurveyCollection

log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs')
def sif(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None
    year_specific_by_generic = year_specific_by_generic_data_frame_name(year)

    erfs_survey_collection = SurveyCollection.load(collection = 'erfs', config_files_directory = config_files_directory)
    erfs_survey = erfs_survey_collection.get_survey('erfs_{}'.format(year))

    log.info("05_foyer: extraction des données foyer")
    # TODO Comment choisir le rfr n -2 pour éxonération de TH ?
    # mnrvka  Revenu TH n-2
    # mnrvkh  revenu TH (revenu fiscal de référence)
    #
    # On récupère les variables du code sif
    sif = erfs_survey.get_values(
        variables = ["noindiv", 'sif', "nbptr", "mnrvka", "rbg", "tsrvbg", "declar"],
        table = year_specific_by_generic["foyer"]
        )

    sif['statmarit'] = 0

    if year == 2009:
        old_sif = sif['sif'][sif['noindiv'] == 901803201].copy()
        new_sif = old_sif.str[0:59] + old_sif.str[60:] + "0"
        sif.sif.loc[sif['noindiv'] == 901803201] = new_sif.values
        old_sif = sif.sif.loc[sif['noindiv'] == 900872201]
        new_sif = old_sif.str[0:58] + " " + old_sif.str[58:]
        sif.sif.loc[sif['noindiv'] == 900872201] = new_sif.values
        del old_sif, new_sif

    sif["rbg"] = sif["rbg"] * ((sif["tsrvbg"] == '+').astype(int) - (sif["tsrvbg"] == '-').astype(int))
    sif["stamar"] = sif.sif.str[4:5]

    # Converting marital status
    statmarit_dict = {"M": 1, "C": 2, "D": 3, "V": 4, "O": 5}
    for key, val in statmarit_dict.items():
        sif.statmarit.loc[sif.stamar == key] = val

    sif["birthvous"] = sif.sif.str[5:9]
    sif["birthconj"] = sif.sif.str[10:14]
    sif["caseE"] = sif.sif.str[15:16] == "E"
    sif["caseF"] = sif.sif.str[16:17] == "F"
    sif["caseG"] = sif.sif.str[17:18] == "G"
    sif["caseK"] = sif.sif.str[18:19] == "K"

    d = 0
    if year in [2006, 2007]:
        sif["caseL"] = sif.sif.str[19:20] == "L"
        sif["caseP"] = sif.sif.str[20:21] == "P"
        sif["caseS"] = sif.sif.str[21:22] == "S"
        sif["caseW"] = sif.sif.str[22:23] == "W"
        sif["caseN"] = sif.sif.str[23:24] == "N"
        sif["caseH"] = sif.sif.str[24:28]
        sif["caseT"] = sif.sif.str[28:29] == "T"

    if year in [2008]:
        d = - 1  # fin de la case L
        sif["caseP"] = sif.sif.str[20 + d: 21 + d] == "P"
        sif["caseS"] = sif.sif.str[21 + d: 22 + d] == "S"
        sif["caseW"] = sif.sif.str[22 + d: 23 + d] == "W"
        sif["caseN"] = sif.sif.str[23 + d: 24 + d] == "N"
        sif["caseH"] = sif.sif.str[24 + d: 28 + d]
        sif["caseT"] = sif.sif.str[28 + d: 29 + d] == "T"

    if year in [2009]:
        sif["caseL"] = sif.sif.str[19: 20] == "L"
        sif["caseP"] = sif.sif.str[20: 21] == "P"
        sif["caseS"] = sif.sif.str[21: 22] == "S"
        sif["caseW"] = sif.sif.str[22: 23] == "W"
        sif["caseN"] = sif.sif.str[23: 24] == "N"
        # caseH en moins par rapport à 2008 (mais case en L en plus)
        # donc décalage par rapport à 2006
        d = -4
        sif["caseT"] = sif.sif.str[28 + d: 29 + d] == "T"

    sif["caseX"] = sif.sif.str[33 + d: 34 + d] == "X"
    sif["dateX"] = sif.sif.str[34 + d: 42 + d]
    sif["caseY"] = sif.sif.str[42 + d: 43 + d] == "Y"
    sif["dateY"] = sif.sif.str[43 + d: 51 + d]
    sif["caseZ"] = sif.sif.str[51 + d: 52 + d] == "Z"
    sif["dateZ"] = sif.sif.str[52 + d: 60 + d]
    sif["causeXYZ"] = sif.sif.str[60 + d: 61 + d]

    # TODO: convert dateXYZ to appropriate date in pandas
    # print(sif["dateY"].unique())

    sif["nbptr"] = sif.nbptr.values / 100
    sif["rfr_n_2"] = sif.mnrvka.values

    sif["nbF"] = sif.sif.str[64 + d: 66 + d]
    sif["nbG"] = sif.sif.str[67 + d: 69 + d]
    sif["nbR"] = sif.sif.str[70 + d: 72 + d]
    sif["nbJ"] = sif.sif.str[73 + d: 75 + d]
    sif["nbN"] = sif.sif.str[76 + d: 78 + d]
    sif["nbH"] = sif.sif.str[79 + d: 81 + d]
    sif["nbI"] = sif.sif.str[82 + d: 84 + d]

    if (year != 2009):
        sif["nbP"] = sif.sif.str[85 + d: 87 + d]

    del sif["stamar"]

    duplicated_noindiv = sif.noindiv[sif.noindiv.duplicated()].copy()
    sif['duplicated_noindiv'] = sif.noindiv.isin(duplicated_noindiv)
    x = sif.loc[sif.duplicated_noindiv, ['noindiv', 'declar']]
    sif['change'] = "NONE"
    sif.loc[sif.duplicated_noindiv, 'change'] = sif.loc[sif.duplicated_noindiv, 'declar'].str[27:28]

    log.info("Number of individuals: {}".format(len(sif.noindiv)))
    log.info("Number of duplicated individuals: {}".format(len(duplicated_noindiv)))
    log.info("Number of distinct individuals: {}".format(len(sif.noindiv.value_counts())))

    log.info("Saving sif")
    temporary_store['sif_{}'.format(year)] = sif
    del sif
    gc.collect()


@temporary_store_decorator(file_name = 'erfs')
def foyer_all(temporary_store = None, year = None):
    year_specific_by_generic = year_specific_by_generic_data_frame_name(year)

    # On ajoute les cases de la déclaration
    erfs_survey_collection = SurveyCollection.load(collection = 'erfs', config_files_directory = config_files_directory)
    data = erfs_survey_collection.get_survey('erfs_{}'.format(year))
    foyer_all = data.get_values(table = year_specific_by_generic["foyer"])
    # on ne garde que les cases de la déclaration ('_xzz') ou ^_[0-9][a-z]{2}")
    regex = re.compile("^_[0-9][a-z]{2}")
    variables = [x for x in foyer_all.columns if regex.match(x)]
    # rename variable to fxzz ou ^f[0-9][a-z]{2}")
    renamed_variables = ["f{}".format(x[1:]) for x in variables]

    foyer = foyer_all[variables + ["noindiv"]].copy()  # Memory expensive ...
    del foyer_all
    gc.collect()
    foyer.rename(columns = dict(zip(variables, renamed_variables)), inplace = True)

    # On aggrège les déclarations dans le cas où un individu a fait plusieurs déclarations
    foyer = foyer.groupby("noindiv", as_index = False).aggregate(numpy.sum)
    print_id(foyer)

    # On récupère les variables individualisables
    var_dict = {
        'sali': ['f1aj', 'f1bj', 'f1cj', 'f1dj', 'f1ej'],
        'hsup': ['f1au', 'f1bu', 'f1cu', 'f1du', 'f1eu'],
        'choi': ['f1ap', 'f1bp', 'f1cp', 'f1dp', 'f1ep'],
        'fra': ['f1ak', 'f1bk', 'f1ck', 'f1dk', 'f1ek'],
        'cho_ld': ['f1ai', 'f1bi', 'f1ci', 'f1di', 'f1ei'],
        'ppe_tp_sa': ['f1ax', 'f1bx', 'f1cx', 'f1dx', 'f1qx'],
        'ppe_du_sa': ['f1av', 'f1bv', 'f1cv', 'f1dv', 'f1qv'],
        'rsti': ['f1as', 'f1bs', 'f1cs', 'f1ds', 'f1es'],
        'alr': ['f1ao', 'f1bo', 'f1co', 'f1do', 'f1eo'],
        'f1tv': ['f1tv', 'f1uv'],
        'f1tw': ['f1tw', 'f1uw'],
        'f1tx': ['f1tx', 'f1ux'],
        'ppe_tp_ns': ['f5nw', 'f5ow', 'f5pw'],
        'ppe_du_ns': ['f5nv', 'f5ov', 'f5pv'],
        'frag_exon': ['f5hn', 'f5in', 'f5jn'],
        'frag_impo': ['f5ho', 'f5io', 'f5jo'],
        'arag_exon': ['f5hb', 'f5ib', 'f5jb'],
        'arag_impg': ['f5hc', 'f5ic', 'f5jc'],
        'arag_defi': ['f5hf', 'f5if', 'f5jf'],
        'nrag_exon': ['f5hh', 'f5ih', 'f5jh'],
        'nrag_impg': ['f5hi', 'f5ii', 'f5ji'],
        'nrag_defi': ['f5hl', 'f5il', 'f5jl'],
        'nrag_ajag': ['f5hm', 'f5im', 'f5jm'],
        'mbic_exon': ['f5kn', 'f5ln', 'f5mn'],
        'abic_exon': ['f5kb', 'f5lb', 'f5mb'],
        'nbic_exon': ['f5kh', 'f5lh', 'f5mh'],
        'mbic_impv': ['f5ko', 'f5lo', 'f5mo'],
        'mbic_imps': ['f5kp', 'f5lp', 'f5mp'],
        'abic_impn': ['f5kc', 'f5lc', 'f5mc'],
        'abic_imps': ['f5kd', 'f5ld', 'f5md'],
        'nbic_impn': ['f5ki', 'f5li', 'f5mi'],
        'nbic_imps': ['f5kj', 'f5lj', 'f5mj'],
        'abic_defn': ['f5kf', 'f5lf', 'f5mf'],
        'abic_defs': ['f5kg', 'f5lg', 'f5mg'],
        'nbic_defn': ['f5kl', 'f5ll', 'f5ml'],
        'nbic_defs': ['f5km', 'f5lm', 'f5mm'],
        'nbic_apch': ['f5ks', 'f5ls', 'f5ms'],
        'macc_exon': ['f5nn', 'f5on', 'f5pn'],
        'aacc_exon': ['f5nb', 'f5ob', 'f5pb'],
        'nacc_exon': ['f5nh', 'f5oh', 'f5ph'],
        'macc_impv': ['f5no', 'f5oo', 'f5po'],
        'macc_imps': ['f5np', 'f5op', 'f5pp'],
        'aacc_impn': ['f5nc', 'f5oc', 'f5pc'],
        'aacc_imps': ['f5nd', 'f5od', 'f5pd'],
        'aacc_defn': ['f5nf', 'f5of', 'f5pf'],
        'aacc_defs': ['f5ng', 'f5og', 'f5pg'],
        'nacc_impn': ['f5ni', 'f5oi', 'f5pi'],
        'nacc_imps': ['f5nj', 'f5oj', 'f5pj'],
        'nacc_defn': ['f5nl', 'f5ol', 'f5pl'],
        'nacc_defs': ['f5nm', 'f5om', 'f5pm'],
        'mncn_impo': ['f5ku', 'f5lu', 'f5mu'],
        'cncn_bene': ['f5sn', 'f5ns', 'f5os'],
        'cncn_defi': ['f5sp', 'f5nu', 'f5ou', 'f5sr'], # TODO: check
        'mbnc_exon': ['f5hp', 'f5ip', 'f5jp'],
        'abnc_exon': ['f5qb', 'f5rb', 'f5sb'],
        'nbnc_exon': ['f5qh', 'f5rh', 'f5sh'],
        'mbnc_impo': ['f5hq', 'f5iq', 'f5jq'],
        'abnc_impo': ['f5qc', 'f5rc', 'f5sc'],
        'abnc_defi': ['f5qe', 'f5re', 'f5se'],
        'nbnc_impo': ['f5qi', 'f5ri', 'f5si'],
        'nbnc_defi': ['f5qk', 'f5rk', 'f5sk'],
        # 'ebic_impv' : ['f5ta','f5ua', 'f5va'],
        # 'ebic_imps' : ['f5tb','f5ub', 'f5vb'],
        'mbic_mvct': ['f5hu'],
        'macc_mvct': ['f5iu'],
        'mncn_mvct': ['f5ju'],
        'mbnc_mvct': ['f5kz'],
        'frag_pvct': ['f5hw', 'f5iw', 'f5jw'],
        'mbic_pvct': ['f5kx', 'f5lx', 'f5mx'],
        'macc_pvct': ['f5nx', 'f5ox', 'f5px'],
        'mbnc_pvct': ['f5hv', 'f5iv', 'f5jv'],
        'mncn_pvct': ['f5ky', 'f5ly', 'f5my'],
        'mbic_mvlt': ['f5kr', 'f5lr', 'f5mr'],
        'macc_mvlt': ['f5nr', 'f5or', 'f5pr'],
        'mncn_mvlt': ['f5kw', 'f5lw', 'f5mw'],
        'mbnc_mvlt': ['f5hs', 'f5is', 'f5js'],
        'frag_pvce': ['f5hx', 'f5ix', 'f5jx'],
        'arag_pvce': ['f5he', 'f5ie', 'f5je'],
        'nrag_pvce': ['f5hk', 'f5lk', 'f5jk'],
        'mbic_pvce': ['f5kq', 'f5lq', 'f5mq'],
        'abic_pvce': ['f5ke', 'f5le', 'f5me'],
        'nbic_pvce': ['f5kk', 'f5ik', 'f5mk'],
        'macc_pvce': ['f5nq', 'f5oq', 'f5pq'],
        'aacc_pvce': ['f5ne', 'f5oe', 'f5pe'],
        'nacc_pvce': ['f5nk', 'f5ok', 'f5pk'],
        'mncn_pvce': ['f5kv', 'f5lv', 'f5mv'],
        'cncn_pvce': ['f5so', 'f5nt', 'f5ot'],
        'mbnc_pvce': ['f5hr', 'f5ir', 'f5jr'],
        'abnc_pvce': ['f5qd', 'f5rd', 'f5sd'],
        'nbnc_pvce': ['f5qj', 'f5rj', 'f5sj'],
        'demenage': ['f1ar', 'f1br', 'f1cr', 'f1dr', 'f1er'],  # (déménagement) uniquement en 2006
        }
    cases_f6_f7_f8 = build_cerfa_fields_by_column_name(year = year, sections_cerfa = [6, 7, 8])
    var_dict.update(cases_f6_f7_f8)
    vars_sets = [set(var_list) for var_list in list(var_dict.values())]
    eligible_vars = (set().union(*vars_sets)).intersection(set(list(foyer.columns)))

    log.info(
        "From {} variables, we keep {} eligibles variables".format(
            len(set().union(*vars_sets)),
            len(eligible_vars),
            )
        )

    qui = ['vous', 'conj', 'pac1', 'pac2', 'pac3']
    #    err = 0
    #    err_vars = {}

    foy_ind = DataFrame()
    for individual_var, foyer_vars in var_dict.items():
        try:
            selection = foyer[foyer_vars + ["noindiv"]].copy()
        except KeyError:
            # Testing if at least one variable of foyers_vars is in the eligible list
            presence = [x in eligible_vars for x in foyer_vars]
            if not any(presence):
                log.info("{} is not present".format(individual_var))
                continue
            else:
                # Shrink the list
                foyer_vars_cleaned = [var for var, present in zip(foyer_vars, presence) if present is True]
                selection = foyer[foyer_vars_cleaned + ["noindiv"]].copy()

        # Reshape the dataframe
        selection.rename(columns = dict(zip(foyer_vars, qui)), inplace = True)
        selection.set_index("noindiv", inplace = True)
        selection.columns.name = "quifoy"

        selection = selection.stack()
        selection.name = individual_var
        selection = selection.reset_index()  # A Series cannot see its index resetted to produce a DataFrame
        selection = selection.set_index(["quifoy", "noindiv"])
        selection = selection[selection[individual_var] != 0].copy()

        if len(foy_ind) == 0:
            foy_ind = selection
        else:
            foy_ind = concat([foy_ind, selection], axis = 1, join = 'outer')

    foy_ind.reset_index(inplace = True)

    ind_vars_to_remove = Series(list(eligible_vars))
    temporary_store['ind_vars_to_remove_{}'.format(year)] = ind_vars_to_remove
    foy_ind.rename(columns = {"noindiv": "idfoy"}, inplace = True)

    print_id(foy_ind)

    foy_ind.quifoy.loc[foy_ind.quifoy == 'vous'] = 0
    foy_ind.quifoy.loc[foy_ind.quifoy == 'conj'] = 1
    foy_ind.quifoy.loc[foy_ind.quifoy == 'pac1'] = 2
    foy_ind.quifoy.loc[foy_ind.quifoy == 'pac2'] = 3
    foy_ind.quifoy.loc[foy_ind.quifoy == 'pac3'] = 4

    assert foy_ind.quifoy .isin(range(5)).all(), 'présence de valeurs aberrantes dans quifoy'

    log.info('saving foy_ind')
    print_id(foy_ind)
    temporary_store['foy_ind_{}'.format(year)] = foy_ind

    return


if __name__ == '__main__':
    year = 2009
    # import sys
    # logging.basicConfig(level = logging.INFO, stream = sys.stdout)
    logging.basicConfig(level = logging.INFO,  filename = 'step_05.log', filemode = 'w')
    sif(year = year)
    foyer_all(year = year)
    log.info("étape 05 foyer terminée")



"""ERFS-FPR aggregates integrations tests."""

import configparser
import click
import logging
import numpy as np
import pandas as pd
import sys
import gc
import os



from openfisca_france_data import france_data_tax_benefit_system
from openfisca_france_data.erfs_fpr import REFERENCE_YEAR
from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario
from openfisca_france_data.aggregates import FranceAggregates as Aggregates


log = logging.getLogger(__name__)
logging.basicConfig(level = logging.INFO, stream = sys.stdout,
    format='%(asctime)s - %(name)-12s: %(levelname)s %(module)s - %(funcName)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)


def test_erfs_fpr_survey_simulation_aggregates(year = REFERENCE_YEAR, rebuild_input_data = False, use_marginal_tax_rate = True, variation_factor = 0.03, varying_variable = 'salaire_de_base'):
    log.info(f'test_erfs_fpr_survey_simulation_aggregates for {year}...')
    np.seterr(all = 'raise')
    tax_benefit_system = france_data_tax_benefit_system
    survey_name = 'openfisca_erfs_fpr_' + str(year)
    survey_scenario = get_survey_scenario(
        tax_benefit_system = tax_benefit_system,
        year = year,
        rebuild_input_data = rebuild_input_data,
        use_marginal_tax_rate = use_marginal_tax_rate,
        variation_factor = variation_factor,
        varying_variable = varying_variable,
        survey_name = survey_name,
        )
    aggregates_taxipp = Aggregates(survey_scenario = survey_scenario, target_source = 'taxipp')
    aggregates_ines = Aggregates(survey_scenario = survey_scenario, target_source = 'ines')

    if False:
        mtr_rd = survey_scenario.compute_marginal_tax_rate(target_variable = 'revenu_disponible', period = year, use_baseline = True)
        print("Rev Disp: Mean = {}; Zero = {}; Positive = {}; Total = {};".format(mtr_rd.mean(), sum(mtr_rd == 0), sum(mtr_rd > 0), mtr_rd.size))
        np.quantile(mtr_rd, q = np.arange(0, 1.1, .1))

        vv1 = survey_scenario.simulation.calculate_add('salaire_de_base', period = year)
        vv2 = survey_scenario._modified_simulation.calculate_add('salaire_de_base', period = year)

        tv1 = survey_scenario.simulation.calculate_add('revenu_disponible', period = year)
        tv2 = survey_scenario._modified_simulation.calculate_add('revenu_disponible', period = year)

        np.quantile(mtr_rd, q = np.arange(0, 1.1, .1))

    return survey_scenario, aggregates_taxipp.get_data_frame(), aggregates_ines.get_data_frame()


def test_erfs_fpr_aggregates_reform():
    """Tests aggregates value with data.

    :param year: year of data and simulation to test agregates
    :param reform: optional argument, put an openfisca_france.refoms object, default None
    """
    tax_benefit_system = france_data_tax_benefit_system
    year = 2014
    survey_scenario = get_survey_scenario(
        reform = 'plf2015',
        baseline_tax_benefit_system = tax_benefit_system,
        year = year,
        rebuild_input_data = False,
        )
    aggregates = Aggregates(survey_scenario = survey_scenario)
    base_data_frame = aggregates.compute_aggregates()
    difference_data_frame = aggregates.compute_difference()

    return aggregates, base_data_frame, difference_data_frame


@click.command()
@click.option('-y', '--year', 'year', default = REFERENCE_YEAR, help = "ERFS-FPR year", show_default = True,
    type = int, required = True)
@click.option('-c', '--configfile', default = None,
    help = 'raw_data.ini path to read years to process.', show_default = True)
@click.option('-v', '--verbose', default = False,
    help = 'print debug information', show_default = True)
def main(year, configfile = None, verbose = False):
    """Computes aggregates."""
    if verbose:
        logging.basicConfig(level = logging.DEBUG, stream = sys.stdout)

    # marginal tax rate parameters
    varying_variable = 'salaire_de_base'
    target_variable = 'revenu_disponible'
    relative_variation = 0.03

    years = []
    if configfile is not None:
        log.warning(f"Reading years to process from {configfile}, ignoring 'year' input parameter")
        try:
            config = configparser.ConfigParser()
            config.read(configfile)
            for key in config['erfs_fpr']:
                if key.isnumeric():
                    years.append(int(key))
                    log.info(f"Adding year {int(key)}")
        except KeyError:
            years = [year]
            log.warning(f"Key 'erfs_fpr' not found in {configfile}, switching back to year {year}")
    else:
        years = [year]

    for year in years:
        survey_scenario, aggregates_taxipp, aggregates_ines = test_erfs_fpr_survey_simulation_aggregates(
            year = year,
            rebuild_input_data = False,
            use_marginal_tax_rate = True,
            variation_factor = relative_variation,
            varying_variable = varying_variable
            )

        df = pd.concat({
            "taxipp": aggregates_taxipp,
            "ines": aggregates_ines,
            })
        df.to_csv(f'aggregates_erfs_fpr_{year}.csv')

        continue
        survey_scenario._set_used_as_input_variables_by_entity()

        mtr_rd = survey_scenario.compute_marginal_tax_rate(target_variable = target_variable, period = year, use_baseline = True)
        print("Rev Disp: Mean = {}; Zero = {}; Positive = {}; Total = {};".format(mtr_rd.mean(), sum(mtr_rd == 0), sum(mtr_rd > 0), mtr_rd.size))
        gc.collect()

        var_level = "b+1"

        # salaire_de_base ne bouge pas :
        # ppe, rev_cap, pens_nettes
        # exclues dans les décompositions suivantes, pour l'instant

        if var_level == "basic":
            vars_to_export = [
                'salaire_de_base',
                'revenu_disponible',
                'niveau_de_vie',
                'revenus_nets_du_travail',
                'revenus_nets_du_capital',
                'pensions_nettes',
                'impots_directs',
                'prestations_sociales',
                # 'ppe',
                'wprm',
                ]
        elif var_level == "b+1":
            vars_to_export = [
                'salaire_de_base',
                'revenu_disponible',
                'niveau_de_vie',
                'revenus_nets_du_travail',
                'salaire_net',
                'rpns_imposables',
                'csg_imposable_non_salarie',
                'crds_non_salarie',
                # 'revenus_nets_du_capital',
                # 'pensions_nettes',
                'impots_directs',
                'taxe_habitation',
                'irpp_economique',
                'prelevement_forfaitaire_liberatoire',
                'prelevement_forfaitaire_unique_ir',
                'ir_pv_immo',
                'isf_ifi',
                'prestations_sociales',
                'prestations_familiales',
                'minima_sociaux',
                'aides_logement',
                'reduction_loyer_solidarite',
                'covid_aide_exceptionnelle_famille_montant',
                'covid_aide_exceptionnelle_tpe_montant',
                # 'ppe'
                'wprm',
                ]

        print("Computing baseline data frame")
        # dtbl = survey_scenario.create_data_frame_by_entity(vars_to_export, index=True, use_modified=False, merge=True)
        dtbl = survey_scenario.create_data_frame_by_entity(vars_to_export, index=True, merge=True)
        print("Saving to disk")
        dtbl.to_csv(f"dt_baseline_erfs_fpr_{year}.csv")
        gc.collect()

        print("Computing reform data frame")
        # dtrf = survey_scenario.create_data_frame_by_entity(vars_to_export, index=True, use_modified=True, merge=True)
        dtrf = survey_scenario.create_data_frame_by_entity(vars_to_export, index=True, merge=True)
        print("Saving to disk")
        dtrf.to_csv(f"dt_reform_erfs_fpr_{year}.csv")
        gc.collect()

        print("Launching R script..")

        # 'vsc' option necessary to indicate right path to R; the number afterwards is the individual ID for the cas types
        # Note BCO : this code won't work outside a specific env
        # os.system('echo 0070 | sudo -S Rscript ~/Analysis/Debug/MTR-Components_Python.R vsc 42 {} {} {}'.format(varying_variable, target_variable, relative_variation))

        print("All done!")


if __name__ == '__main__':
    log.info("Starting...")
    main()
    log.info("THE END!")



import os
import pkg_resources
import sys
import gc

from pandas import HDFStore
from openfisca_france.utils import check_consistency
#    Uses rpy2.
#    On MS Windows, The environment variable R_HOME and R_USER should be set

try:
    import pandas.rpy.common as com
    import rpy2.rpy_classic as rpy
    rpy.set_default_mode(rpy.NO_CONVERSION)
except Exception:
    pass


#from openfisca_france.data.sources.config import DATA_DIR

openfisca_france_location = pkg_resources.get_distribution('openfisca-france-data').location
CONFIG_DIR = os.path.join(openfisca_france_location)

#ERF_HDF5_DATA_DIR = os.path.join(SRC_PATH,'countries','france','data', 'erf')


class Survey(object):
    """
    An object to describe survey data
    """
    def __init__(self, name = None, label = None, **kwargs):

        assert name is not None, "A survey should have a name"
        self.name = name
        self.tables = dict()
        if label is None:
            self.label = self.name
        else:
            self.label = label
        self.informations = kwarg


    def insert_table(self, name=None, **kwargs):
        """
        Insert a table in the Survey
        """
        if name not in self.tables.keys():
            self.tables[name] = dict()

        for key, val in kwargs.items():
            if key in ["RData_dir", "RData_filename", "variables"]:
                    self.tables[name][key] = val


class SurveyCollection(object):
    """
    A collection of Surveys
    """
    def __init__(self, year):

        super(SurveyCollection, self).__init__()
        self.surveys = dict()
        self.hdf5_filename = None


def build_erfs_survey_collection():
        # self.hdf5_filename = os.path.join(os.path.dirname(ERF_HDF5_DATA_DIR),'erf','erf.h5')


    erfs_survey_collection = SurveyCollection()
    for year in range(2006, 2010):

        surveys = erfs_survey_collection.surveys
        yr = str(year)[2:]
        yr1 = str(year+1)[2:]

        eec_variables = ['noi','noicon','noindiv','noiper','noimer','ident','naia','naim','lien',
                       'acteu','stc','contra','titc','mrec','forter','rstg','retrai','lpr','cohab','sexe',
                       'agepr','rga','statut', 'txtppb', 'encadr', 'prosa', 'nbsala',  'chpub', 'dip11']
        eec_rsa_variables = [ "sp0" + str(i) for i in range(0,10)] + ["sp10", "sp11"] + ['sitant', 'adeben',
                            'datant', 'raistp', 'amois', 'adfdap' , 'ancentr', 'ancchom', 'dimtyp', 'rabsp', 'raistp',
                             'rdem', 'ancinatm']
        eec_aah_variables = ["rc1rev", "maahe"]
        eec_variables += eec_rsa_variables + eec_aah_variables

        erf_tables = {
            "erf_menage" : {
                "RData_filename" : "menage" + yr,
                "year" : year,
                "variables" : None,
                },
            "eec_menage" : {
                "RData_filename" : "mrf" + yr + "e" + yr + "t4",
                "year" : year,
                "variables" : None,
                },
            "foyer" : {
                "RData_filename" : "foyer" + yr,
                "year" : year,
                "variables" : None,
                },
            "erf_indivi" : {
                "RData_filename" : "indivi" + yr,
                "year" : year,
                "variables" : ['noi','noindiv','ident','declar1','quelfic','persfip','declar2','persfipd','wprm',
                     "zsali","zchoi","ztsai","zreti","zperi","zrsti","zalri","zrtoi","zragi","zrici","zrnci",
                     "zsalo","zchoo","ztsao","zreto","zpero","zrsto","zalro","zrtoo","zrago","zrico","zrnco",
                     ],
                },
            "eec_indivi" : {
                "RData_filename" : "irf" + yr + "e" + yr + "t4",
                "year" : year,
                "variables" : eec_variables,
                },
            "eec_cmp_1" : {
                "RData_filename" : "icomprf" + yr + "e" + yr1 + "t1",
                "year" : year,
                "variables" : eec_variables,
                },
            "eec_cmp_2" : {
                "RData_filename" : "icomprf" + yr + "e" + yr1 + "t2",
                "year" : year,
                "variables" : eec_variables,
                },
            "eec_cmp_3" : {
                "RData_filename" : "icomprf" + yr + "e" + yr1 + "t3",
                "year" : year,
                "variables" : eec_variables,
                },
        }

        # Build absolute path for RData_filename
        from configparser import SafeConfigParser

        parser = SafeConfigParser()
        config_local_ini = os.path.join(CONFIG_DIR, 'config_local.ini')
        config_ini = os.path.join(CONFIG_DIR, 'config.ini')
        found = parser.read(config_local_ini, config_ini)
        data_directory = parser.get('data', 'input_directory')
        for table in erf_tables:
            table["RData_filename"] = os.path.join(os.path.dirname(data_directory),'R','erf')

    def initialize(self):
        """
        Initialize survey data
        """

        self.initialize_erf(tables=tables)
#        self.initialize_logement()

    def initialize_erf(self, tables=None):
        """
        """

        year = self.year
        erf = SurveyDescription()
        yr = str(year)[2:]
        yr1 = str(year+1)[2:]
        erf_tables_to_process = {
                                 "erf_menage" : "menage" + yr,
                                 "eec_menage" : "mrf" + yr + "e" + yr + "t4",
                                  "foyer" : "foyer" + yr,
                                   "erf_indivi" : "indivi" + yr,
                                "eec_indivi" : "irf" + yr + "e" + yr + "t4",
                                "eec_cmp_1" : "icomprf" + yr + "e" + yr1 + "t1",
                                "eec_cmp_2" : "icomprf" + yr + "e" + yr1 + "t2",
                                "eec_cmp_3" : "icomprf" + yr + "e" + yr1 + "t3"
                                }
        RData_dir = os.path.join(os.path.dirname(DATA_DIR),'R','erf')

        variables = ['noi','noindiv','ident','declar1','quelfic','persfip','declar2','persfipd','wprm',
                     "zsali","zchoi","ztsai","zreti","zperi","zrsti","zalri","zrtoi","zragi","zrici","zrnci",
                     "zsalo","zchoo","ztsao","zreto","zpero","zrsto","zalro","zrtoo","zrago","zrico","zrnco"]

        variables_eec = ['noi','noicon','noindiv','noiper','noimer','ident','naia','naim','lien',
                       'acteu','stc','contra','titc','mrec','forter','rstg','retrai','lpr','cohab','sexe',
                       'agepr','rga','statut', 'txtppb', 'encadr', 'prosa', 'nbsala',  'chpub', 'dip11']

        variables_eec_rsa = [ "sp0" + str(i) for i in range(0,10)] + ["sp10", "sp11"] + ['sitant', 'adeben',
                            'datant', 'raistp', 'amois', 'adfdap' , 'ancentr', 'ancchom', 'dimtyp', 'rabsp', 'raistp',
                             'rdem', 'ancinatm']

        variables_eec_aah = ["rc1rev", "maahe"]

        variables_eec += variables_eec_rsa + variables_eec_aah

        erf_tables = {
            "erf_menage" : {"RData_filename" :  "menage" + yr,
                            "variables" : None},
            "eec_menage" : {"RData_filename" :"mrf" + yr + "e" + yr + "t4",
                            "variables" : None},
            "foyer" :      {"RData_filename" :"foyer" + yr,
                            "variables" : None},
            "erf_indivi" : {"RData_filename" :"indivi" + yr,
                            "variables" : variables},
            "eec_indivi" : {"RData_filename" :"irf" + yr + "e" + yr + "t4",
                            "variables" : variables_eec},
            "eec_cmp_1" :  {"RData_filename" :"icomprf" + yr + "e" + yr1 + "t1",
                            "variables" : variables_eec},
            "eec_cmp_2" :  {"RData_filename" :"icomprf" + yr + "e" + yr1 + "t2",
                            "variables" : variables_eec},
            "eec_cmp_3" :  {"RData_filename" :"icomprf" + yr + "e" + yr1 + "t3",
                            "variables" : variables_eec}}

        RData_dir = os.path.join(os.path.dirname(DATA_DIR),'R','erf')

        if tables is None:
            erf_tables_to_process = erf_tables
        else:
            erf_tables_to_process = tables

        for name in erf_tables_to_process:
            erf.insert_table(name=name,
                             RData_filename=RData_filename,
                             RData_dir=RData_dir,
                             variables=variables)

        self.surveys["erf"] = erf


    def initialize_logement(self):
        """
        """
        year = self.year
        lgt = SurveyDescription()
        yr = str(year)[2:]
        yr1 = str(year+1)[2:]


        if yr=="03":
            lgt_men = "menage"
            lgt_logt = None
            renameidlgt  = dict(ident='ident')

        elif yr in ["06","07","08","09"]:
            lgt_men = "menage1"
            lgt_lgt = "logement"
            renameidlgt = dict(idlog='ident')

        lgt_tables_to_process = {"adresse" : "adresse",
                                 "lgt_menage" : lgt_men,
                                 "lgt_logt" : lgt_lgt}

        RData_dir = os.path.join(os.path.dirname(DATA_DIR),'R','logement')
        for name, RData_filename in lgt_tables_to_process.items():
            lgt.insert_table(name=name,
                             RData_filename=RData_filename,
                             RData_dir=RData_dir)

        self.surveys["lgt"] = lgt


    def initialize_patrimoine(self, year):
        """
        TODO:
        """
        pat_tables_to_process = {"pat_individu" : "individu",
                                 "pat_menage" : "meange",
                                 "pat_produit" : "produit",
                                 "pat_transmission" : "transm"}

        pat_data_dir = os.path.join(os.path.dirname(DATA_DIR),'R','patrimoine')

        pat = {"name" : "patrimoine",
               "data_dir" : os.path.join(os.path.dirname(DATA_DIR),'R','patrimoine'),
               "tables_to_process" : pat_tables_to_process}



    def set_config(self, **kwargs):
        """
        Set configuration parameters

        Parameters
        ----------
        year : int, default None
               year of the survey
        """
        if self.year is not None:
            year = self.year
        else:
            raise Exception("year should be defined")

        store = HDFStore(self.hdf5_filename)
        for survey_name, description in self.surveys.items():
            for destination_table_name, tables in description.tables.items():
                data_dir = tables["RData_dir"]
                R_table_name = tables["RData_filename"]
                try:
                    variables = tables["variables"]
                except Exception:
                    variables = None
                print(variables)
                self.store_survey(survey_name, R_table_name, destination_table_name, data_dir, variables)

    def store_survey(self, survey_name, R_table_name, destination_table_name, data_dir, variables=None, force_recreation=True):
        """
        Store a R data table in an HDF5 file

        Parameters
        ----------

        survey_name : string
                       the name of the survey
        R_table_name : string
                       the name of the R data table
        destination_table_name : string
                                 the name of the table in the HDFStore
        data_dir : path
                   the directory where to find the RData file

        variables : list of string, default None
                    When not None, list of the variables to keep
        """
        gc.collect()
        year = self.year
        def get_survey_year(survey_name, year):
            if survey_name == "logement":
                if year == 2003:
                    return 2003
                elif year in range(2006,2010):
                    return 2006
            if survey_name == "patrimoine":
                return 2004
            else:
                return year

        print("creating %s" %(destination_table_name))
        table_Rdata = R_table_name + ".Rdata"
        filename = os.path.join(data_dir, str(get_survey_year(survey_name, year)), table_Rdata)
        print(filename)
        if not os.path.isfile(filename):
            raise Exception("filename do  not exists")

        rpy.r.load(filename)
        stored_table = com.load_data(R_table_name)
        store = HDFStore(self.hdf5_filename)
        store_path = str(self.year)+"/"+destination_table_name

        if store_path in store:
            if force_recreation is not True:
                print(store_path + "already exists, do not re-create and exit")
                store.close()
                return

        if variables is not None:

            print(store)
            print(store_path)
            print(variables)
            variables_stored = list(set(variables).intersection(set(stored_table.columns)))
            print(list(set(variables).difference((set(stored_table.columns)))))
            store[store_path] = stored_table[variables_stored]
        else:
            store[store_path] = stored_table
        store.close()
        del stored_table
        gc.collect()


    def get_value(self, variable, table=None):
        """
        Get value

        Parameters
        ----------
        variable : string
                  name of the variable
        table : string, default None
                name of the table where to get variable
        Returns
        -------
        df : DataFrame, default None
             A DataFrame containing the variable
        """
        df = self.get_values([variable], table)
        return df

    def get_values(self, variables=None, table=None):
        """
        Get values

        Parameters
        ----------
        variables : list of strings, default None
                  list of variables names, if None return the whole table
        table : string, default None
                name of the table where to get the variables
        Returns
        -------
        df : DataFrame, default None
             A DataFrame containing the variables
        """

        store = HDFStore(self.hdf5_filename)
        df = store[str(self.year)+"/"+table]
        # If no variables read the whole table
        if variables is None:
            return df

        diff = set(variables) - set(df.columns)
        if diff:
            raise Exception("The following variable(s) %s are missing" %diff)
        variables = list( set(variables).intersection(df.columns))
        df = df[variables]

        return df


    def get_of_value(self, variable, table=None):
        """
        Get value

        Parameters
        ----------
        variable : string
                  name of the variable
        table : string, default None
                name of the table where to get variable
        Returns
        -------
        df : DataFrame, default None
             A DataFrame containing the variable
        """
        df = self.get_of_values([variable], table)
        return df


    def get_of_values(self, variables=None, table=None):
        """
        Get values

        Parameters
        ----------
        variables : list of strings, default None
                  list of variables names, if None return the whole table
        table : string, default None
                name of the table where to get the variables
        Returns
        -------
        df : DataFrame, default None
             A DataFrame containing the variables
        """

        store = HDFStore(self.hdf5_filename)
        df = store[str(self.year)+"/"+table]

        # If no variables read the whole table
        if variables is None:
            return df

        from openfisca_france.data.erf import get_erf2of, get_of2erf
        of2erf = get_of2erf()
        to_be_renamed_variables = set(of2erf.keys()).intersection(variables)
        renamed_variables = []


        for variable in to_be_renamed_variables:
            renamed_variables.append(of2erf[variable])

        if renamed_variables:
            variables = list( set(variables).difference(to_be_renamed_variables)) + renamed_variables

#        if table is None:
#            for test_table in self.tables.keys:
#                if set(variables) < set(self.tables[test_table].columns):
#                    table = test_table
#                    print("using guessed table :", table)
#                    break
#
#        if table is None:
#            print("varname not found in any tables")
#            df = None
#        else:

        variables = list( set(variables).intersection(df.columns))
        df = df[variables]

        # rename variables according to their name in openfisca
        erf2of = get_erf2of()
        to_be_renamed_variables = set(list(erf2of.values())).intersection(variables)
        if to_be_renamed_variables:
            for var in to_be_renamed_variables:
                df.rename(columns = {var: erf2of[var]}, inplace=True)
        return df

## def test():
##     '''
##     Validate check_consistency
##     '''
##     #===========================================================================
##     # from pandas import DataFrame
##     #res = DataFrame({af_col.name: simulation.output_table.get_value(af_col.name, af_col.entity)})
##     # print(res)
##     #===========================================================================

##     store = HDFStore(os.path.join(os.path.dirname(os.path.join(SRC_PATH,'countries','france','data','erf')),'fichiertest.h5'))
##     datatable = store.get('test12')
##     test_simu = store.get('test_simu')
##     print(check_consistency(test_simu, datatable))

## def test3():
##     year=2006
##     erf = SurveyCollection(year=year)
##     df = erf.get_of_values(table = "eec_menage")
##     from openfisca_core.simulations import SurveySimulation
##     simulation = SurveySimulation()
##     simulation.set_config(year=year)
##     simulation.set_param()
##     simulation.compute() # TODO: this should not be mandatory
##     check_consistency(simulation.input_table, df)

## def test_init():

##     for year in range(2009,2010):
##         data = SurveyCollection(year=year)
##         data.initialize(tables=["eec_indivi"])
##         data.set_config()

## #def test_reading_stata_tables():
## #    from pandas.io.stata import StataReader, read_stata # TODO: wait for the next release ...
## #
## #    filename = os.path.join(DATA_DIR,"erf","2006","Tables complémentaires","icomprf06e07t1.dta")
## #    reader = StataReader(filename)
## #    print(reader.data())



if __name__ == '__main__':
#     test3()
    test_init()
    ## hdf5_filename = os.path.join(os.path.dirname(ERF_HDF5_DATA_DIR),'erf','erf.h5')
    ## print(hdf5_filename)
    ## store = HDFStore(hdf5_filename)
    ## print(store)

    build_erfs_survey_collection()

#
#     hdf5_filename = os.path.join(os.path.dirname(ERF_HDF5_DATA_DIR),'erf','erf_old.h5')
#     print(hdf5_filename)
#     store = HDFStore(hdf5_filename)
#     print(store)



"""Configuration file parser."""

import configparser
from pathlib import Path, PurePath
from xdg import BaseDirectory


default_config_files_directory = Path(BaseDirectory.save_config_path('openfisca-france-data'))


class Config(configparser.SafeConfigParser):
    """Custom Config Parser."""

    config_ini = None

    def __init__(self, config_files_directory = default_config_files_directory):
        configparser.SafeConfigParser.__init__(self)
        assert config_files_directory is not None

        config_ini = PurePath.joinpath(config_files_directory, 'config.ini')
        if Path.exists(config_ini):
            self.config_ini = config_ini
        self.read([config_ini])

    def save(self):
        """Save config to home user config for package openfisca-france-data."""
        assert self.config_ini, "configuration file paths are not defined"
        config_file = open(self.config_ini, 'w')
        self.write(config_file)
        config_file.close()


config = Config()



#! /usr/bin/env python
import logging


from openfisca_france_data.utils import control, print_id
from openfisca_survey_manager.temporary import temporary_store_decorator


log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs')
def invalide(temporary_store = None, year = None):

    assert temporary_store is not None
    assert year is not None

    log.info("Entering 07_invalides: construction de la variable invalide")
# # # Invalides
# # #invalide = caseP (vous), caseF (conj) ou case G, caseI, ou caseR (pac)

# # loadTmp("final.Rdata")
# # invalides <- final[,c("noindiv","idmen","caseP","caseF","idfoy","quifoy")]
# # invalides <- within(invalides,{
# #   caseP <- ifelse(is.na(caseP),0,caseP)
# #   caseF <- ifelse(is.na(caseF),0,caseF)
# #   invalide <- FALSE})
# # # Les "vous" invalides
# # table(invalides[,c("caseF","quifoy")],useNA="ifany")
# # invalides[(invalides$caseP==1) & (invalides$quifoy=="vous"),"invalide"] <- TRUE
# #

    log.info("Etape 1 : création de la df invalides")
    log.info("    1.1 : déclarants invalides")
    final = temporary_store['final_{}'.format(year)]
    if "invalide" in final:
        # on drop les colonnes inv et alt au cas ou on aurait déjà lancé le step07
        final.drop(["invalide", "alt"], axis = 1, inplace = True)

    invalides_vars = [
        "caseF",
        "caseP",
        "idfoy",
        "idmen",
        "noindiv",
        "quifoy",
        ]

    aah_eec_variables = ["rc1rev", "maahe"]
    aah_eec = False

    if set(aah_eec_variables) < set(final.columns):
        aah_eec = True
        invalides_vars += aah_eec_variables

    assert set(invalides_vars) < set(final.columns), \
        "Variables {} are missing".format(set(invalides_vars).difference(set(final.columns)))

    invalides = final.xs(invalides_vars, axis = 1).copy()

    invalides['invalide'] = False
    invalides['alt'] = False
    for var in ["caseP", "caseF"]:
        assert invalides[var].notnull().all(), 'NaN values in {}'.format(var)

    # Les déclarants invalides
    invalides.loc[(invalides['caseP'] == 1) & (invalides['quifoy'] == 0), 'invalide'] = True
    log.info("Il y a {} invalides déclarants".format(invalides["invalide"].sum()))

    # Les personnes qui touchent l'aah dans l'enquête emploi
    if aah_eec:
        log.info("Inspecting rc1rev")
        log.info(invalides['rc1rev'].value_counts())
        invalides.loc[invalides.maahe > 0, 'invalide'] = True
        invalides.loc[invalides.rc1rev == 4, 'invalide'] = True  # TODO: vérifier le format.
#  TODO:      invalides.rc1rev.astype("str") voir mai mahdi pour pendre en compte 14 24 etc

        log.info("Il y a {} invalides qui touchent des alloc".format(invalides["invalide"].sum()))

    print_id(invalides)

# # # Les conjoints invalides
# #
# # #men_inv_conj <- invalides[c("idmen","caseF","quifoy")]
# # #men_inv_conj <- rename(men_inv_conj, c("caseF"="invalide"))
# # #table(men_inv_conj[men_inv_conj$inv==1 ,c("invalide","quifoy")],useNA="ifany")
# # # Il y a des caseF suir des conjoints cela vint des doubles d?clarations TODO: shoumd clean this
# # #toto <- invalides[invalides$caseF==1 & invalides$quifoy=="conj","idmen"]
# # #load(indm)
# # #titi <- indivim[(indivim$ident %in% toto) & (indivim$persfip=="vous" |indivim$persfip=="conj") ,c("ident","noindiv","declar1","declar2","persfip","quelfic")]
# # #titi <- titi[order(titi$ident),]
# # foy_inv_conj <- invalides[,c("idfoy","caseF","quifoy")]
# # foy_inv_conj <- rename(foy_inv_conj, c("caseF"="invalide"))
# # table(foy_inv_conj[ ,c("invalide","quifoy")],useNA="ifany")
# # # On ne garde donc que les caseF des "vous"

# # foy_inv_conj   <- foy_inv_conj[foy_inv_conj$quifoy=="vous",c("idfoy","invalide")]
# # table(foy_inv_conj[ ,c("invalide")],useNA="ifany")
# # invalides_conj <- invalides[invalides$quifoy=="conj",c("idfoy","noindiv")]
# # invalides_conj <- merge(invalides_conj, foy_inv_conj, by="idfoy", all.x=TRUE)
# # table(invalides_conj$inv) # TODO en 2006 On en a 316 au lieu de 328 il doit y avoir de idfoy avec caseF qui n'ont pas de vous because double déclaration'
# # invalides[invalides$quifoy=="conj",c("idfoy","noindiv","invalide")] <- invalides_conj
# # table(invalides[,c("invalide","quifoy")],useNA="ifany")
# # rm(invalides_conj,foy_inv_conj)

    # On récupère les idfoy des foyers avec une caseF cochée
    log.info('    1.2 : Les conjoints invalides')
    idfoy_inv_conj = final.idfoy[final.caseF].copy()
    inv_conj_condition = (invalides.idfoy.isin(idfoy_inv_conj) & (invalides.quifoy == 1))
    invalides.loc[inv_conj_condition, "invalide"] = True

    log.info("Il y a {} invalides conjoints".format(len(invalides[inv_conj_condition])))
    log.info(" Il y a {} invalides déclarants et invalides conjoints".format(invalides["invalide"].sum()))

    # Enfants invalides et garde alternée
    # #
    # # loadTmp("pacIndiv.Rdata")
    # # foy_inv_pac <- invalides[!(invalides$quifoy %in% c("vous","conj")),c("invalide","noindiv")]
    # # foy_inv_pac <- merge(foy_inv_pac, pacIndiv[,c("noindiv","typ","naia")], by="noindiv",all.x =TRUE)
    # # names(foy_inv_pac)
    # # table(foy_inv_pac[,c("typ","naia")],useNA="ifany")
    # # table(foy_inv_pac[,c("typ")],useNA="ifany")
    # # foy_inv_pac <- within(foy_inv_pac,{
    # #   invalide  <- (typ=="G") | (typ=="R") | (typ=="I") | (typ=="F" & (as.numeric(year)-naia>18))
    # #   alt  <- (typ=="H") | (typ=="I")
    # #   naia <- NULL
    # #   typ  <- NULL})
    # #
    # # table(foy_inv_pac[ ,c("invalide")],useNA="ifany")
    # # table(foy_inv_pac[ ,c("alt")],useNA="ifany")
    # # invalides$alt <- 0
    # # foy_inv_pac[is.na(foy_inv_pac$alt),"alt"] <- 0
    # # invalides[!(invalides$quifoy %in% c("vous","conj")),c("noindiv","invalide","alt")] <- foy_inv_pac
    log.info("    1.3 : enfants invalides et en garde alternée (variables inv et alt)")
    pacIndiv = temporary_store['pacIndiv_{}'.format(year)]
    log.info(pacIndiv.type_pac.value_counts())

    foy_inv_pac = invalides[['noindiv', 'invalide']][~(invalides.quifoy.isin([0, 1]))].copy()
    # pac = pacIndiv.ix[:, ["noindiv", "type_pac", "naia"]]
    log.info("{}".format(len(foy_inv_pac)))

    log.info("{}".format(pacIndiv.columns))
    foy_inv_pac = foy_inv_pac.merge(
        pacIndiv[['noindiv', 'type_pac', 'naia']].copy(),
        on = 'noindiv',
        how = 'left',
        )
    foy_inv_pac['invalide'] = (
        foy_inv_pac['type_pac'].isin(['G', 'R', 'I']) |
        (
            (foy_inv_pac.type_pac == "F") & ((year - foy_inv_pac.naia) > 18)
            )
        )
    foy_inv_pac['alt'] = ((foy_inv_pac.type_pac == "H") | (foy_inv_pac.type_pac == "I"))
    del foy_inv_pac['naia']
    del foy_inv_pac['type_pac']
    foy_inv_pac['alt'] = foy_inv_pac['alt'].fillna(False)

    log.info("{}".format(foy_inv_pac['invalide'].describe()))
    invalides.loc[:, 'alt'] = False
    invalides.loc[~(invalides.quifoy.isin([0, 1])), ["alt", "invalide"]] = foy_inv_pac[["alt", "invalide"]].copy().values
    invalides = invalides[["noindiv", "invalide", 'alt']].copy()
    invalides['alt'].fillna(False, inplace = True)

    log.info(invalides.invalide.value_counts())
    # invalides = invalides.drop_duplicates(['noindiv', 'invalide', 'alt'], take_last = True)
    del foy_inv_pac, pacIndiv

    # Initialisation des NA sur alt et inv
    # invalides[is.na(invalides$inv), "invalide"] <- 0
    # table(invalides[,c("alt","invalide")],useNA="ifany")
    #
    # final <- merge(final, invalides[,c("noindiv","invalide","alt")], by="noindiv",all.x=TRUE)
    # table(final[, c("invalide","alt")],useNA="ifany")

    log.info('Etape 2 : Initialisation des NA sur alt et inv')
    assert invalides.invalide.notnull().all() & invalides.alt.notnull().all()
    final.set_index('noindiv', inplace = True, verify_integrity = True)
    invalides.set_index('noindiv', inplace = True, verify_integrity = True)
    final = final.join(invalides)
    final.reset_index(inplace = True)
    del invalides

    log.info("{}".format(final.invalide.value_counts()))
    control(final, debug = True)

    temporary_store['final_{}'.format(year)] = final
    log.info(u'final complétée et sauvegardée')

if __name__ == '__main__':
    logging.basicConfig(level = logging.INFO, filename = 'step_07.log', filemode = 'w')
    year = 2009
    invalide(year = year)
    log.info("étape 07 création des invalides terminée")



import pytest

from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario as erfs_fpr_get_survey_scenario
from openfisca_france_data.erfs.scenario import ErfsSurveyScenario
from openfisca_france_data import base_survey


def test_pivot_table_1d_mean():
    year = 2013
    survey_scenario = get_survey_scenario(kind = 'erfs_fpr', year = year)
    pivot_table = survey_scenario.compute_pivot_table(
        columns = ['decile_rfr'],
        values = ['irpp'],
        period = year,
        )
    return pivot_table


def test_pivot_table_1d_sum():
    year = 2013
    survey_scenario = get_survey_scenario(kind = 'erfs_fpr', year = year)
    pivot_table = survey_scenario.compute_pivot_table(
        aggfunc = 'sum',
        columns = ['decile_rfr'],
        values = ['irpp'],
        period = year,
        )
    return pivot_table


def test_pivot_table_1d_count():
    year = 2013
    survey_scenario = get_survey_scenario(kind = 'erfs_fpr', year = year)
    pivot_table = survey_scenario.compute_pivot_table(
        aggfunc = 'count',
        columns = ['decile_rfr'],
        values = ['irpp'],
        period = year,
        )
    return pivot_table


def test_pivot_table_2d_2values():
    year = 2013
    survey_scenario = get_survey_scenario(kind = 'erfs_fpr', year = year)
    pivot_table = survey_scenario.compute_pivot_table(
        columns = ['decile_rfr'],
        index = ['nbptr'],
        values = ['irpp', 'rfr'],
        period = year,
        )
    return pivot_table


def get_survey_scenario(kind = 'erfs_fpr', year = None):
    if kind == 'erfs_fpr':
        tax_benefit_system = base_survey.france_data_tax_benefit_system
        return erfs_fpr_get_survey_scenario(year = year, tax_benefit_system = tax_benefit_system)
    else:
        year = 2009
        SurveyScenario = ErfsSurveyScenario

    return SurveyScenario.create(year = year)



import logging
import numpy as np

from openfisca_core import periods


log = logging.getLogger(__name__)


def create_taux_csg_remplacement(individus, period, tax_benefit_system, sigma = (28.1) ** 2):
    """
    """
    assert 'revkire' in individus
    assert 'nbp' in individus
    assert period.start.year >= 2015

    rfr = individus.revkire
    nbptr = individus.nbp / 100

    def compute_taux_csg_remplacement(rfr, nbptr):
        parameters = tax_benefit_system.get_parameters_at_instant(period.start)
        seuils = parameters.prelevements_sociaux.contributions_sociales.csg.remplacement.pensions_de_retraite_et_d_invalidite
        seuil_exoneration = seuils.seuil_de_rfr_1 + (nbptr - 1) * seuils.demi_part_suppl
        seuil_reduction = seuils.seuil_de_rfr_2 + (nbptr - 1) * seuils.demi_part_suppl
        taux_csg_remplacement = 0.0 * rfr
        if period.start.year >= 2019:
            seuil_taux_intermediaire = seuils.seuil_rfr3 + (nbptr - 1) * seuils.demi_part_suppl_rfr3
            taux_csg_remplacement = np.where(
                rfr <= seuil_exoneration,
                1,
                np.where(
                    rfr <= seuil_reduction,
                    2,
                    np.where(
                        rfr <= seuil_taux_intermediaire,
                        3,
                        4,
                        )
                    )
                )
        else:
            taux_csg_remplacement = np.where(
                rfr <= seuil_exoneration,
                1,
                np.where(
                    rfr <= seuil_reduction,
                    2,
                    3,
                    )
                )

        return taux_csg_remplacement

    individus['taux_csg_remplacement'] = compute_taux_csg_remplacement(rfr, nbptr)

    assert individus['taux_csg_remplacement'].isin(range(5)).all()



from openfisca_france_data.felin.input_data_builder.create_variables_individuelles import create_taux_csg_remplacement
from openfisca_france_data.common import (
    create_salaire_de_base,
    create_traitement_indiciaire_brut,
    create_revenus_remplacement_bruts,
    )


def create_individu_variables_brutes(
    individus,
    revenu_type = None,
    period = None,
    tax_benefit_system = None,
    ):
    """
    Crée les variables brutes de revenus:
      - salaire_de_base
      - traitement_indiciaire_brut
      - primes_fonction_publique
      - retraite_bruite
      - chomage_brut
    à partir des valeurs nettes ou imposables de ces revenus
    et d'autres information individuelles
    """
    assert revenu_type in ['imposable', 'net']
    assert period is not None
    assert tax_benefit_system is not None

    assert 'age' in individus.columns
    assert 'contrat_de_travail' in individus.columns
    assert 'heures_remunerees_volume' in individus.columns
    assert 'categorie_salarie' in individus.columns
    assert 'categorie_non_salarie' in individus.columns
    assert 'effectif_entreprise' in individus.columns

    # Temp

    created_variables = []

    create_salaire_de_base(individus, period = period, revenu_type = revenu_type, tax_benefit_system = tax_benefit_system)
    created_variables.append('salaire_de_base')

    create_traitement_indiciaire_brut(individus, period = period, revenu_type = revenu_type, tax_benefit_system = tax_benefit_system)
    created_variables.append('traitement_indiciaire_brut')
    created_variables.append('primes_fonction_publique')

    create_taux_csg_remplacement(individus, period, tax_benefit_system)
    created_variables.append('taux_csg_remplacement')

    create_revenus_remplacement_bruts(individus, period, tax_benefit_system)
    created_variables.append('chomage_brut')
    created_variables.append('retraite_brute')

    return created_variables



import logging


from openfisca_survey_manager.surveys import SurveyCollection
from openfisca_france_data.erf import get_of2erf, get_erf2of
import numpy as np


log = logging.getLogger(__name__)


def build_erf_aggregates(variables = None, year = 2006, unit = 1e6):
    """
    Fetch the relevant aggregates from erf data
    """
    erfs_survey_collection = SurveyCollection.load(collection = "erfs")
    erfs_survey = erfs_survey_collection.surveys["erfs_{}".format(year)]


    of2erf = get_of2erf()
    erf2of = get_erf2of()

    if set(variables) <= set(of2erf.keys()):
        variables = [ of2erf[variable] for variable in variables]

    if variables is not None and "wprm" not in variables:
        variables.append("wprm")
    log.info("Fetching aggregates from erfs {} data".format(year))


    df = erfs_survey.get_values(variables = variables, table = "erf_menage")


    df.rename(columns = erf2of, inplace = True)
    wprm = df["wprm"]
    for col in df.columns:
        try:
            df[col] = df[col].astype(np.float64)
        except Exception:
            pass
    df = df.mul(wprm, axis = 0)
    for col in list(set(df.columns) - set(['ident', 'wprm'])):
        try:
            df[col] = df[col].sum()/1e6
        except Exception:
            pass

    return df.ix[0:1] # Aggregate so we only need 1 row


if __name__ == '__main__':
    df = build_erf_aggregates(variables = ["af"])
    print(df.to_string())



"""Constructin des variables ménages."""


import logging
import pandas as pd

from openfisca_survey_manager.temporary import temporary_store_decorator


log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs_fpr')
def build_variables_menage(temporary_store = None, year = None):
    menages = temporary_store[f'menages_{year}']

    if "loyer" in menages.columns:
        menages['loyer'] = menages['loyer'] * 12

    if year == 2018:
        menages['statut_occupation_logement'] = menages['so'].copy()

    if year == 2019:  # SO ne fonctionne pas en 2019 bcp de 0 = non-renseigné
        menages['statut_occupation_logement'] = menages['logt'].copy()

    if year >= 2018:
        menages['zone_apl'] = 3
        menages.loc[menages['tau2010'] == 10,'zone_apl'] = 1
        menages.loc[menages['tau2010'] == 9,'zone_apl'] = 2
        #menages.loc[menages['tau2010'] == 8,'zone_apl'] = 2
        # pour l'instant on met tout le monde à 2 mais à améliorer, peut être en fonction de la taille de l'aire urbaine ?
        menages.loc[(menages.statut_occupation_logement == 7), 'statut_occupation_logement'] = 2
        menages['logement_conventionne'] = False
        menages.loc[menages['statut_occupation_logement'] == 3 ,'logement_conventionne'] = True
        temporary_store['menages_{}'.format(year)] = menages




import pickle
import csv

if __name__ == '__main__':

    codeDict = {}
    fileName = 'zone_apl.csv'

    code = csv.reader(open(fileName), delimiter = ";")

    for row in code:
        codeDict.update({row[2]: (row[1], row[4])})

    outputFile = open("code_apl", 'wb')
    pickle.dump(codeDict, outputFile)
    outputFile.close()



import logging

from openfisca_france_data import france_data_tax_benefit_system
from openfisca_france_data.erfs_fpr.get_survey_scenario import get_survey_scenario


log = logging.getLogger(__name__)

year = 2014
tax_benefit_system = france_data_tax_benefit_system

survey_scenario = get_survey_scenario(
    tax_benefit_system = tax_benefit_system,
    baseline_tax_benefit_system = tax_benefit_system,
    year = year,
    rebuild_input_data = False,
    )


checks_by_variable = {
    'salaire_de_base': {
        'min': 500e9
        }
    }


for variable in survey_scenario.used_as_input_variables:
    checks = checks_by_variable.get(variable)
    print(variable)
    minimum = checks.get('min') if checks is not None else None
    if minimum:
        assert survey_scenario.compute_aggregate(variable, period = 2014) >= minimum
    survey_scenario.summarize_variable(variable)




from setuptools import setup, find_namespace_packages


with open('README.md') as file:
    long_description = file.read()

setup(
    name = "OpenFisca-France-Data",
    version = "2.0.5",
    description = "OpenFisca-France-Data module to work with French survey data",
    long_description = long_description,
    long_description_content_type="text/markdown",
    author = "OpenFisca Team",
    author_email = "contact@openfisca.fr",
    url = "https://github.com/openfisca/openfisca-france-data",
    license = "http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    keywords = "tax benefit social survey data microsimulation",
    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Information Analysis",
        ],
    package_data = {
        'openfisca_france_data': ['assets/aggregats/taxipp/agregats_tests_taxipp_2_0.xlsx',
                                  'assets/aggregats/ines/ines_2019.json'],
        },
    entry_points = {
        'console_scripts': [
            'build-erfs-fpr=openfisca_france_data.erfs_fpr.input_data_builder:main',
            'compare-erfs-fpr-input=openfisca_france_data.erfs_fpr.comparison:compare',
            'create-test-erfs-fpr=openfisca_france_data.erfs_fpr.test_case_creation:create_test',
            ],
        },
    python_requires = ">=3.9",
    install_requires = [
        "multipledispatch >=0.6.0, <1.0.0",
        "openFisca-france >=150.0.0, <151.0.0",
        "openFisca-survey-manager >=1, <2.0.0",
        "wquantiles >=0.3.0, <1.0.0",  # To compute weighted quantiles
        ],
    extras_require = {
        "test": [
            "click >=8.0.0, <9.0.0",
            "autopep8 >=2.0.2, <3",
            "bumpver >=2022.1120",
            "dtale",
            "flake8 >=6.0.0, <7.0.0",
            "ipdb >=0.13, <1.0",
            "ipython >=7.5.0, <8.0.0",
            "mypy >=0.670, <1.0.0",
            "pypandoc",
            'pytest >=7.2.2, <8.0',
            "scipy >=1.2.1, <2.0.0",
            "toolz >=0.9.0, <1.0.0",
            ],
        },
    packages = find_namespace_packages(exclude = ("docs", "tests")),
    )



#! /usr/bin/env python
import gc
import logging

from pandas import Series, concat
import numpy as np
from numpy import where

from openfisca_survey_manager.temporary import temporary_store_decorator

from openfisca_france_data.utils import print_id, control


log = logging.getLogger(__name__)


@temporary_store_decorator(file_name = 'erfs')
def create_totals_first_pass(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None

    # On part de la table individu de l'ERFS
    # on renomme les variables
    log.info("Creating Totals")
    log.info("Etape 1 : Chargement des données")

    indivim = temporary_store['indivim_{}'.format(year)]

    assert not indivim.duplicated(['noindiv']).any(), "Présence de doublons"

    # Deals individuals with imputed income : some individuals are in 'erf individu table' but
    # not in the 'foyer' table. We need to create a foyer for them.
    selection = Series()
    for var_i in ["zsali", "zchoi", "zrsti", "zalri", "zrtoi", "zragi", "zrici", "zrnci"]:
        var_o = var_i[:-1] + "o"
        test = indivim[var_i] != indivim[var_o]
        if selection.empty:
            selection = test
        else:
            selection = test | selection

    indivi_i = indivim[selection].copy()
    indivi_i.rename(
        columns = {
            "ident": "idmen",
            "persfip": "quifoy",
            "zsali": "sali",  # Inclu les salaires non imposables des agents d'assurance
            "zchoi": "choi",
            "zrsti": "rsti",
            "zalri": "alr"
            },
        inplace = True,
        )
    assert indivi_i.quifoy.notnull().all()
    indivi_i.loc[indivi_i.quifoy == "", "quifoy"] = "vous"
    indivi_i.quelfic = "FIP_IMP"

   # We merge them with the other individuals
    indivim.rename(
        columns = dict(
            ident = "idmen",
            persfip = "quifoy",
            zsali = "sali",  # Inclu les salaires non imposables des agents d'assurance
            zchoi = "choi",
            zrsti = "rsti",
            zalri = "alr",
            ),
        inplace = True,
        )

    if not (set(list(indivim.noindiv)) > set(list(indivi_i.noindiv))):
        raise Exception("Individual ")
    indivim.set_index("noindiv", inplace = True, verify_integrity = True)
    indivi_i.set_index("noindiv", inplace = True, verify_integrity = True)
    indivi = indivim
    del indivim
    indivi.update(indivi_i)
    indivi.reset_index(inplace = True)
    assert not(indivi.noindiv.duplicated().any()), "Doublons"

    log.info("Etape 2 : isolation des FIP")
    fip_imp = indivi.quelfic == "FIP_IMP"
    indivi["idfoy"] = (
        indivi.idmen.astype('int') * 100 +
        (indivi.declar1.str[0:2]).convert_objects(convert_numeric=True)
        )

    # indivi.loc[fip_imp, "idfoy"] = np.nan
    # Certains FIP (ou du moins avec revenus imputés) ont un numéro de déclaration d'impôt ( pourquoi ?)
    assert indivi_i.declar1.notnull().all()
    assert (indivi_i.declar1 == "").sum() > 0
    fip_has_declar = (fip_imp) & (indivi.declar1 != "")

    indivi.loc[fip_has_declar, "idfoy"] = (
        indivi.idmen * 100 + indivi.declar1.str[0:2].convert_objects(convert_numeric = True))
    del fip_has_declar

    fip_no_declar = (fip_imp) & (indivi.declar1 == "")
    del fip_imp
    indivi.loc[fip_no_declar, "idfoy"] = 100 * indivi.loc[fip_no_declar, "idmen"] + indivi.loc[fip_no_declar, "noi"]
    # WAS indivi["idmen"] * 100 + 50

    indivi_fnd = indivi.loc[fip_no_declar, ["idfoy", "noindiv"]].copy()

    while any(indivi_fnd.duplicated(subset = ["idfoy"])):
        indivi_fnd["idfoy"] = where(
            indivi_fnd.duplicated(subset = ["idfoy"]),
            indivi_fnd["idfoy"] + 1,
            indivi_fnd["idfoy"]
            )

    # assert indivi_fnd["idfoy"].duplicated().value_counts()[False] == len(indivi_fnd["idfoy"].values), \
    #    "Duplicates remaining"
    assert not(indivi.noindiv.duplicated().any()), "Doublons"

    indivi.idfoy.loc[fip_no_declar] = indivi_fnd.idfoy.copy()
    del indivi_fnd, fip_no_declar

    log.info("Etape 3 : Récupération des EE_NRT")
    nrt = indivi.quelfic == "EE_NRT"
    indivi.loc[nrt, 'idfoy'] = indivi.loc[nrt, 'idmen'] * 100 + indivi.loc[nrt, 'noi']
    indivi.loc[nrt, 'quifoy'] = "vous"
    del nrt

    pref_or_cref = indivi.lpr.isin([1, 2])
    adults = (indivi.quelfic.isin(["EE", "EE_CAF"])) & (pref_or_cref)
    pref = adults & (indivi.lpr == 1)
    cref = adults & (indivi.lpr == 2)
    indivi.loc[adults, "idfoy"] = indivi.loc[adults, 'idmen'] * 100 + indivi.loc[adults, 'noiprm']
    indivi.loc[pref, "quifoy"] = "vous"
    indivi.loc[cref, "quifoy"] = "conj"
    del adults, pref, cref
    assert indivi.idfoy[indivi.lpr.dropna().isin([1, 2])].all()

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)
    log.info("Il reste {} idfoy problématiques".format(
        indivi_without_idfoy.sum()
        ))

    # Conjoints qui dont il n'existe pas de vous avec leur idfoy.
    # Problème de double déclaration: si leur conjoint à un noindiv à leur idfoy on switch les déclarants
    log.info("{} conj without a valid idfoy".format(
        (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['conj'])).sum()
        ))

    if (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['conj'])).any():
        # On traite les gens qui ont quifoy=conj mais dont l'idfoy n'a pas de vous
        # 1) s'ils ont un conjoint et qu'il est vous avec un idfoy valide on leur attribue son idfoy:
        avec_conjoint = (
            indivi_without_idfoy &
            indivi.idfoy.notnull() &
            indivi.quifoy.isin(['conj']) &
            (indivi.noicon != 0) &
            (100 * indivi.idmen + indivi.noicon).isin(idfoyList)
            )
        indivi.loc[avec_conjoint, 'idfoy'] = (
            100 * indivi.loc[avec_conjoint, 'idmen'] + indivi.loc[avec_conjoint, 'noicon']
            )
        indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)  # mise à jour des cas problématiques
        del avec_conjoint

    if (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['conj'])).any():
        # 2) sinon ils deviennent vous
        devient_vous = (
            indivi_without_idfoy &
            indivi.idfoy.notnull() &
            indivi.quifoy.isin(['conj']) &
            (indivi.noicon == 0)
            )
        indivi.loc[devient_vous, 'idfoy'] = indivi.loc[devient_vous, 'noindiv'].copy()
        indivi.loc[devient_vous, 'quifoy'] = 'vous'
        idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
        indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)  # Mise à jour des cas problématiques
        del devient_vous

    problem = (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['conj']))
    if problem.sum() > 0:
        log.info("Dropping {} conj without valid idfoy".format(
            problem.sum()
        ))
        indivi.drop(indivi[problem].index, inplace = True)
        indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)  # Mise à jour des cas problématiques
        problem = (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['conj']))
        assert not problem.any()

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)
    log.info("Remainning {} non valid idfoy".format(
        indivi_without_idfoy.sum()
        ))
    # Les personnes à charge de personnes repérées comme conjoint idfoy = conj_noindiv
    # (problème des doubles déclarations) doivent récupérer l'idfoy de celles-ci
    pac = (
        indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['pac'])
        )
    log.info("Dealing with {} non valid idfoy of pacs".format(
        pac.sum()
        ))
    conj_noindiv = indivi.idfoy[pac].copy()
    new_idfoy_by_old = indivi.loc[
        indivi.noindiv.isin(conj_noindiv), ['noindiv', 'idfoy']
        ].astype('int').set_index('noindiv').squeeze().to_dict()
    indivi.loc[pac, 'idfoy'] = indivi.loc[pac, 'idfoy'].map(new_idfoy_by_old)
    del pac
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)  # Mise à jour des cas problématiques
    assert not (indivi_without_idfoy & indivi.idfoy.notnull() & indivi.quifoy.isin(['pac'])).any()

    # Il faut traiter les idfoy non attribués
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)
    assert (indivi_without_idfoy == indivi.idfoy.isnull()).all()
    log.info("Il faut traiter les {} idfoy non attribués".format(
        indivi_without_idfoy.sum()
        ))

    # Adultes non enfants avec conjoints déclarants
    married_adult_with_vous = (
        indivi_without_idfoy &
        ((indivi.noiper == 0) | (indivi.noimer == 0)) &
        (indivi.age >= 25) &
        (indivi.noicon > 0) &
        (100 * indivi.idmen + indivi.noicon).isin(idfoyList)
        )
    indivi.loc[married_adult_with_vous, 'idfoy'] = (
        100 * indivi.loc[married_adult_with_vous, 'idmen'] + indivi.loc[married_adult_with_vous, 'noicon']
        )
    indivi.loc[married_adult_with_vous, 'quifoy'] = 'conj'
    log.info(
        """Il y a {} adultes > 25 ans non enfants avec conjoints déclarants""".format(
            married_adult_with_vous.sum()
            )
        )

    # Les deux membres du couples n'ont pas d'idfoy
    married_adult_without_vous = (
        indivi_without_idfoy &
        ((indivi.noiper == 0) | (indivi.noimer == 0)) &
        (indivi.age >= 18) &
        (indivi.noicon > 0) &
        (~married_adult_with_vous)
        )
    # On les groupes par ménages, on vérifie qu'ils ne sont que deux
    couple_by_idmen = (
        (indivi.loc[
            married_adult_without_vous, ['idmen', 'noindiv']
            ].groupby('idmen').agg('count')) == 2).astype('int').squeeze().to_dict()

    couple_idmens = list(idmen for idmen in couple_by_idmen.keys() if couple_by_idmen[idmen])
    # On crée un foyer vous-conj si couple
    vous = married_adult_without_vous & (
        ((indivi.sexe == 1) & indivi.idmen.isin(couple_idmens)) |
        (~indivi.idmen.isin(couple_idmens))
        )
    conj = married_adult_without_vous & (~vous) & indivi.idmen.isin(couple_idmens)
    indivi.loc[vous, 'idfoy'] = indivi.loc[vous, 'noindiv'].copy()
    indivi.loc[vous, 'quifoy'] = 'vous'
    indivi.loc[conj, 'idfoy'] = 100 * indivi.loc[conj, 'idmen'] + indivi.loc[conj, 'noicon']
    indivi.loc[conj, 'quifoy'] = 'conj'
    del vous, conj
    log.info(
        """Il y a {} adultes > 25 ans non enfants sans conjoints déclarants: on crée un foyer""".format(
            married_adult_without_vous.sum()
            )
        )

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    # Cas des enfants agés sans conjoint >= 25 ans
    non_married_aged_kids = (
        indivi_without_idfoy &
        ((indivi.noiper > 0) | (indivi.noimer > 0)) &
        (indivi.age >= 25) &
        (indivi.noicon == 0)
        )
    indivi.loc[non_married_aged_kids, 'idfoy'] = indivi.loc[non_married_aged_kids, 'noindiv'].copy()
    indivi.loc[non_married_aged_kids, 'quifoy'] = 'vous'
    log.info(
        """On crée un foyer fiscal indépendants pour les {} enfants agés de plus de 25 ans sans conjoint
vivant avec leurs parents""".format(
            non_married_aged_kids.sum()
            )
        )
    del non_married_aged_kids
    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    # Cas des enfants agés avec conjoint >= 18 ans
    married_aged_kids = (
        indivi_without_idfoy &
        ((indivi.noiper > 0) | (indivi.noimer > 0)) &
        (indivi.age >= 18) &
        (indivi.noicon != 0)
        )

    # Cas des enfants agés avec conjoint >= 18 ans
    married_aged_kids = (
        indivi_without_idfoy &
        ((indivi.noiper > 0) | (indivi.noimer > 0)) &
        (indivi.age >= 18) &
        (indivi.noicon != 0)
        )
    noiconjs = 100 * indivi.idmen + indivi.noicon
    quifoy_by_noiconj = indivi.loc[
        indivi.noindiv.isin(noiconjs[married_aged_kids]), ['noindiv', 'quifoy']
        ].set_index('noindiv').dropna().squeeze().to_dict()
    is_conj_vous = noiconjs.map(quifoy_by_noiconj) == "vous"
    indivi.loc[married_aged_kids & is_conj_vous, 'quifoy'] = "conj"
    indivi.loc[married_aged_kids & is_conj_vous, 'idfoy'] = noiconjs[married_aged_kids & is_conj_vous].copy()

    log.info("""Il y a {} enfants agés de plus de 25 ans avec conjoint
vivant avec leurs parents qui ne sont pas traités""".format(
        married_aged_kids.sum()
        ))  # Il n'y en a pas en 2009
    del married_aged_kids, noiconjs, is_conj_vous

    # Colocations
    if indivi_without_idfoy.any():
        potential_idmens = indivi.loc[indivi_without_idfoy, 'idmen'].copy()
        colocs = indivi.loc[indivi.idmen.isin(potential_idmens), ['idmen', 'age', 'quifoy']].copy()
        coloc_by_idmen = colocs.groupby('idmen').agg({
            'age':
                lambda x:
                    (abs((x.min() - x.max())) < 20) & (x.min() >= 18),
            'quifoy':
                lambda x:
                    (x == 'vous').sum() >= 1,
                }
            )
        coloc_dummy_by_idmen = (coloc_by_idmen.age * coloc_by_idmen.quifoy)
        coloc_idmens = coloc_dummy_by_idmen.index[coloc_dummy_by_idmen.astype('bool')].tolist()
        colocataires = indivi_without_idfoy & indivi.idmen.isin(coloc_idmens)
        indivi.loc[colocataires, 'quifoy'] = 'vous'
        indivi.loc[colocataires, 'idfoy'] = indivi.loc[colocataires, 'noindiv'].copy()
        log.info("Il y a {} colocataires".format(
            colocataires.sum()
            ))
        del colocataires, coloc_dummy_by_idmen, coloc_by_idmen, coloc_idmens, colocs

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    # On met le reste des adultes de plus de 25 ans dans des foyers uniques
    other_adults = indivi_without_idfoy & (indivi.age >= 25)
    if indivi_without_idfoy.any():
        indivi.loc[other_adults, 'quifoy'] = 'vous'
        indivi.loc[other_adults, 'idfoy'] = indivi.loc[other_adults, 'noindiv'].copy()
        log.info("Il y a {} autres adultes seuls à qui l'on crée un foyer individuel".format(
            other_adults.sum()
            ))
        del other_adults

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    # Cas des enfants jeunes < 25 ans
    kids = (
        indivi_without_idfoy &
        (indivi.age < 25) &
        ((indivi.noiper > 0) | (indivi.noimer > 0))
        )
    # On rattache les enfants au foyer de leur pères s'il existe
    log.info("On traite le cas des {} enfants (noiper ou noimer non nuls) repérés non rattachés".format(
        kids.sum()
        ))
    if kids.any():
        pere_declarant_potentiel = kids & (indivi.noiper > 0)
        indivi['pere_noindiv'] = (100 * indivi.idmen.fillna(0) + indivi.noiper.fillna(0)).astype('int')
        pere_noindiv = (
            100 * indivi.loc[pere_declarant_potentiel, 'idmen'].fillna(0) +
            indivi.loc[pere_declarant_potentiel, 'noiper'].fillna(0)
            ).astype('int')
        idfoy_by_noindiv = indivi.loc[
            indivi.noindiv.isin(pere_noindiv), ['noindiv', 'idfoy']
            ].dropna().astype('int').set_index('noindiv').squeeze().to_dict()
        pere_declarant_potentiel_idfoy = indivi['pere_noindiv'].map(idfoy_by_noindiv)
        pere_veritable_declarant = pere_declarant_potentiel & pere_declarant_potentiel_idfoy.isin(idfoyList)
        indivi.loc[pere_veritable_declarant, 'idfoy'] = (
            pere_declarant_potentiel_idfoy[pere_veritable_declarant].astype('int')
            )
        indivi.loc[pere_veritable_declarant, 'quifoy'] = 'pac'
    log.info("{} enfants rattachés au père ".format(
        pere_veritable_declarant.sum()
        ))
    del pere_declarant_potentiel, pere_declarant_potentiel_idfoy, pere_noindiv, \
        pere_veritable_declarant, idfoy_by_noindiv

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)
    kids = (
        indivi_without_idfoy &
        (indivi.age < 25) &
        ((indivi.noiper > 0) | (indivi.noimer > 0))
        )
    log.info("Il reste {} enfants (noimer non nuls) repérés non rattachés".format(
        kids.sum()
        ))
    # Et de leurs mères sinon
    if kids.any():
        mere_declarant_potentiel = kids & (indivi.noimer > 0)
        indivi['mere_noindiv'] = (100 * indivi.idmen.fillna(0) + indivi.noimer.fillna(0)).astype('int')
        mere_noindiv = (
            100 * indivi.loc[mere_declarant_potentiel, 'idmen'].fillna(0) +
            indivi.loc[mere_declarant_potentiel, 'noimer'].fillna(0)
            ).astype('int')
        idfoy_by_noindiv = indivi.loc[
            indivi.noindiv.isin(mere_noindiv), ['noindiv', 'idfoy']
            ].dropna().astype('int').set_index('noindiv').squeeze().to_dict()
        mere_declarant_potentiel_idfoy = indivi['mere_noindiv'].map(idfoy_by_noindiv)
        mere_veritable_declarant = mere_declarant_potentiel & mere_declarant_potentiel_idfoy.isin(idfoyList)
        indivi.loc[mere_veritable_declarant, 'idfoy'] = (
            mere_declarant_potentiel_idfoy[mere_veritable_declarant].astype('int')
            )
        indivi.loc[mere_veritable_declarant, 'quifoy'] = 'pac'
    log.info("{} enfants rattachés à la mère".format(
        mere_veritable_declarant.sum()
        ))
    del mere_declarant_potentiel, mere_declarant_potentiel_idfoy, mere_noindiv, \
        mere_veritable_declarant, idfoy_by_noindiv

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    # Enfants avec parents pas indiqués (noimer et noiper = 0)
    if indivi_without_idfoy.any():
        potential_idmens = indivi.loc[indivi_without_idfoy, 'idmen'].copy()
        parents = indivi.loc[indivi.idmen.isin(potential_idmens), [
            'idmen', 'age', 'quifoy', 'lpr', 'noiper', 'noimer']].copy()
        parents_by_idmen = parents.groupby('idmen').agg({
            'quifoy':
                lambda quifoy: (quifoy == 'vous').sum() == 1,
            }
        )
        parents_dummy_by_idmen = parents_by_idmen.quifoy.copy()
        parents_idmens = parents_dummy_by_idmen.index[
            parents_dummy_by_idmen.astype('bool')].tolist()
        parents_idfoy_by_idmem = indivi.loc[
            indivi.idmen.isin(parents_idmens) & (indivi.quifoy == 'vous'),
            ['idmen', 'noindiv']].dropna().astype('int').set_index('idmen').squeeze().to_dict()
        avec_parents = (
            indivi_without_idfoy &
            indivi.idmen.isin(parents_idmens) &
            (
                (indivi.age < 18) |
                (
                    (indivi.age < 25) &
                    (indivi.sali == 0) &
                    (indivi.choi == 0) &
                    (indivi.alr == 0)
                    )
                ) &
            (indivi.lpr == 4) &
            (indivi.noiper == 0) &
            (indivi.noimer == 0) &
            (indivi.lpr == 4)
            )
        indivi.loc[avec_parents, 'idfoy'] = (
            indivi.loc[avec_parents, 'idmen'].map(parents_idfoy_by_idmem))
        indivi.loc[avec_parents, 'quifoy'] = 'pac'

        log.info("Il y a {} enfants sans noiper ni noimer avec le seul vous du ménage".format(
            avec_parents.sum()
            ))
        del parents, parents_by_idmen, parents_dummy_by_idmen, parents_idfoy_by_idmem, parents_idmens

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    if indivi_without_idfoy.any():
        potential_idmens = indivi.loc[indivi_without_idfoy, 'idmen'].copy()
        parents_non_pr = indivi.loc[
            indivi.idmen.isin(potential_idmens) & (indivi.quifoy == 'vous'),
            ['idmen', 'quifoy', 'noindiv', 'lpr']].copy()
        parents_by_idmen = parents_non_pr.groupby('idmen').filter(
            lambda df: (
                ((df.quifoy == 'vous').sum() >= 1) &
                (df.lpr > 2).any()
            )).query('lpr > 2')
        parents_idfoy_by_idmem = parents_by_idmen[
            ['idmen', 'noindiv']
            ].dropna().astype('int').set_index('idmen').squeeze().to_dict()
        avec_parents_non_pr = (
            indivi_without_idfoy &
            indivi.idmen.isin(parents_idfoy_by_idmem.keys()) &
            (indivi.age < 18) &
            (indivi.lpr == 4) &
            (indivi.noiper == 0) &
            (indivi.noimer == 0)
            )
        indivi.loc[avec_parents_non_pr, 'idfoy'] = (
            indivi.loc[avec_parents_non_pr, 'idmen'].map(parents_idfoy_by_idmem))
        indivi.loc[avec_parents_non_pr, 'quifoy'] = 'pac'

        log.info("Il y a {} enfants sans noiper ni noimer avec le seul vous du ménage".format(
            avec_parents_non_pr.sum()
            ))
        del parents_non_pr, parents_by_idmen, parents_idfoy_by_idmem, avec_parents_non_pr

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    if indivi_without_idfoy.any():
        other_enfants = indivi_without_idfoy & (indivi.age < 18)
        potential_idmens = indivi.loc[other_enfants, 'idmen'].copy()
        declarants = indivi.loc[
            indivi.idmen.isin(potential_idmens) & (indivi.quifoy == 'vous'),
            ['idmen', 'idfoy']].dropna().astype('int').copy()
        declarants_by_idmen = declarants.groupby('idmen').agg({
            'idfoy': 'max'
            }).squeeze().to_dict()
        indivi.loc[other_enfants, 'idfoy'] = indivi.loc[other_enfants, 'idmen'].copy().map(declarants_by_idmen)
        indivi.loc[other_enfants, 'quifoy'] = 'pac'

        log.info("Il y a {} autres enfants que l'on met avec un vous du ménage".format(
            other_enfants.sum()
            ))

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    if indivi_without_idfoy.any():
        other_grands_enfants = indivi_without_idfoy & (indivi.age >= 18)
        indivi.loc[other_grands_enfants, 'idfoy'] = indivi.loc[other_grands_enfants, 'noindiv']
        indivi.loc[other_grands_enfants, 'quifoy'] = 'vous'

        log.info("Il y a {} autres grans enfants (>= 18) que l'on met avec un vous du ménage".format(
            other_grands_enfants.sum()
            ))

    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    temporary_store['indivi_step_06_{}'.format(year)] = indivi

    assert not indivi_without_idfoy.any()

    log.info("    4.2 : On enlève les individus pour lesquels il manque le déclarant")
    fip = temporary_store['fipDat_{}'.format(year)]
    fip["declar"] = np.nan
    fip["agepf"] = np.nan

    fip.drop(["actrec", "year", "noidec"], axis = 1, inplace = True)
    fip.naia = fip.naia.astype("int32")
    fip.rename(
        columns = dict(
            ident = "idmen",
            persfip = "quifoy",
            zsali = "sali",  # Inclu les salaires non imposables des agents d'assurance
            zchoi = "choi",
            zrsti = "rsti",
            zalri = "alr"),
        inplace = True)

    is_fip_19_25 = ((year - fip.naia - 1) >= 19) & ((year - fip.naia - 1) < 25)

    # TODO: BUT for the time being we keep them in thier vous menage so the following lines are commented
    # The idmen are of the form 60XXXX we use idmen 61XXXX, 62XXXX for the idmen of the kids over 18 and less than 25

    indivi = concat([indivi, fip.loc[is_fip_19_25].copy()])
    temporary_store['indivi_step_06_{}'.format(year)] = indivi

    assert not(indivi.noindiv.duplicated().any())

    del is_fip_19_25
    indivi['age'] = year - indivi.naia - 1
    indivi['age_en_mois'] = 12 * indivi.age + 12 - indivi.naim

    indivi["quimen"] = 0
    assert indivi.lpr.notnull().all()
    indivi.loc[indivi.lpr == 1, 'quimen'] = 0
    indivi.loc[indivi.lpr == 2, 'quimen'] = 1
    indivi.loc[indivi.lpr == 3, 'quimen'] = 2
    indivi.loc[indivi.lpr == 4, 'quimen'] = 3
    indivi['not_pr_cpr'] = None  # Create a new row
    indivi.loc[indivi.lpr <= 2, 'not_pr_cpr'] = False
    indivi.loc[indivi.lpr > 2, 'not_pr_cpr'] = True

    assert indivi.not_pr_cpr.isin([True, False]).all()

    log.info("    4.3 : Creating non pr=0 and cpr=1 idmen's")

    indivi.set_index('noindiv', inplace = True, verify_integrity = True)
    test1 = indivi.loc[indivi.not_pr_cpr, ['quimen', 'idmen']].copy()
    test1['quimen'] = 2

    j = 2
    while any(test1.duplicated(['quimen', 'idmen'])):
        test1.loc[test1.duplicated(['quimen', 'idmen']), 'quimen'] = j + 1
        j += 1
    print_id(indivi)
    indivi.update(test1)
    indivi.reset_index(inplace = True)
    print_id(indivi)

    temporary_store['indivi_step_06_{}'.format(year)] = indivi
    gc.collect()
    return


@temporary_store_decorator(file_name = 'erfs')
def create_totals_second_pass(temporary_store = None, year = None):
    assert temporary_store is not None
    assert year is not None

    log.info("    5.1 : Elimination idfoy restant")
    # Voiture balai
    # On a plein d'idfoy vides, on fait 1 ménage = 1 foyer fiscal
    indivi = temporary_store['indivi_step_06_{}'.format(year)]
    idfoyList = indivi.loc[indivi.quifoy == "vous", 'idfoy'].unique()
    indivi_without_idfoy = ~indivi.idfoy.isin(idfoyList)

    indivi.loc[indivi_without_idfoy, 'quifoy'] = "pac"

    indivi.loc[indivi_without_idfoy & (indivi.quimen == 0) & (indivi.age >= 18), 'quifoy'] = "vous"
    indivi.loc[indivi_without_idfoy & (indivi.quimen == 0) & (indivi.age >= 18), 'idfoy'] = (
        indivi.loc[indivi_without_idfoy, "idmen"].astype('int') * 100 + 51
        )
    indivi.loc[indivi_without_idfoy & (indivi.quimen == 1) & (indivi.age >= 18), 'quifoy'] = "conj"

    del idfoyList
    print_id(indivi)

    # Sélectionne les variables à garder pour les steps suivants
    variables = [
        "actrec",
        "age",
        "age_en_mois",
        "chpub",
        "encadr",
        "idfoy",
        "idmen",
        "nbsala",
        "noi",
        "noindiv",
        "prosa",
        "quelfic",
        "quifoy",
        "quimen",
        "statut",
        "titc",
        "txtppb",
        "wprm",
        "rc1rev",
        "maahe",
        "sali",
        "rsti",
        "choi",
        "alr",
        "wprm",
        ]

    assert set(variables).issubset(set(indivi.columns)), \
        "Manquent les colonnes suivantes : {}".format(set(variables).difference(set(indivi.columns)))

    dropped_columns = [variable for variable in indivi.columns if variable not in variables]
    indivi.drop(dropped_columns, axis = 1, inplace = True)

    #  see http://stackoverflow.com/questions/11285613/selecting-columns
    indivi.reset_index(inplace = True)
    gc.collect()

    # TODO les actrec des fip ne sont pas codées (on le fera à la fin quand on aura rassemblé
    # les infos provenant des déclarations)
    log.info("Etape 6 : Création des variables descriptives")
    log.info("    6.1 : Variable activité")
    log.info("Variables présentes; \n {}".format(indivi.columns))

    indivi['activite'] = np.nan
    indivi.loc[indivi.actrec <= 3, 'activite'] = 0
    indivi.loc[indivi.actrec == 4, 'activite'] = 1
    indivi.loc[indivi.actrec == 5, 'activite'] = 2
    indivi.loc[indivi.actrec == 7, 'activite'] = 3
    indivi.loc[indivi.actrec == 8, 'activite'] = 4
    indivi.loc[indivi.age <= 13, 'activite'] = 2  # ce sont en fait les actrec=9
    log.info("Valeurs prises par la variable activité \n {}".format(indivi['activite'].value_counts(dropna = False)))
    # TODO: MBJ problem avec les actrec
    # TODO: FIX AND REMOVE
    indivi.loc[indivi.actrec.isnull(), 'activite'] = 5
    indivi.loc[indivi.titc.isnull(), 'titc'] = 0
    assert indivi.titc.notnull().all(), \
        "Problème avec les titc"  # On a 420 NaN pour les varaibels statut, titc etc

    log.info("    6.2 : Variable statut")
    indivi.loc[indivi.statut.isnull(), 'statut'] = 0
    indivi.statut = indivi.statut.astype('int')
    indivi.loc[indivi.statut == 11, 'statut'] = 1
    indivi.loc[indivi.statut == 12, 'statut'] = 2
    indivi.loc[indivi.statut == 13, 'statut'] = 3
    indivi.loc[indivi.statut == 21, 'statut'] = 4
    indivi.loc[indivi.statut == 22, 'statut'] = 5
    indivi.loc[indivi.statut == 33, 'statut'] = 6
    indivi.loc[indivi.statut == 34, 'statut'] = 7
    indivi.loc[indivi.statut == 35, 'statut'] = 8
    indivi.loc[indivi.statut == 43, 'statut'] = 9
    indivi.loc[indivi.statut == 44, 'statut'] = 10
    indivi.loc[indivi.statut == 45, 'statut'] = 11
    assert indivi.statut.isin(range(12)).all(), "statut value over range"
    log.info("Valeurs prises par la variable statut \n {}".format(
        indivi['statut'].value_counts(dropna = False)))

    log.info("    6.3 : variable txtppb")
    indivi.loc[indivi.txtppb.isnull(), 'txtppb'] = 0
    assert indivi.txtppb.notnull().all()
    indivi.loc[indivi.nbsala.isnull(), 'nbsala'] = 0
    indivi.nbsala = indivi.nbsala.astype('int')
    indivi.loc[indivi.nbsala == 99, 'nbsala'] = 10
    assert indivi.nbsala.isin(range(11)).all()
    log.info("Valeurs prises par la variable txtppb \n {}".format(
        indivi['txtppb'].value_counts(dropna = False)))

    log.info("    6.4 : variable chpub et CSP")
    indivi.loc[indivi.chpub.isnull(), 'chpub'] = 0
    indivi.chpub = indivi.chpub.astype('int')
    assert indivi.chpub.isin(range(11)).all()

    indivi['cadre'] = 0
    indivi.loc[indivi.prosa.isnull(), 'prosa'] = 0
    assert indivi.prosa.notnull().all()
    log.info("Valeurs prises par la variable encadr \n {}".format(indivi['encadr'].value_counts(dropna = False)))

    # encadr : 1=oui, 2=non
    indivi.loc[indivi.encadr.isnull(), 'encadr'] = 2
    indivi.loc[indivi.encadr == 0, 'encadr'] = 2
    assert indivi.encadr.notnull().all()
    assert indivi.encadr.isin([1, 2]).all()

    indivi.loc[indivi.prosa.isin([7, 8]), 'cadre'] = 1
    indivi.loc[(indivi.prosa == 9) & (indivi.encadr == 1), 'cadre'] = 1
    assert indivi.cadre.isin(range(2)).all()

    log.info(
        "Etape 7: on vérifie qu'il ne manque pas d'info sur les liens avec la personne de référence"
        )
    log.info(
        "nb de doublons idfoy/quifoy {}".format(len(indivi[indivi.duplicated(subset = ['idfoy', 'quifoy'])]))
        )

    log.info("On crée les n° de personnes à charge dans le foyer fiscal")
    assert indivi.idfoy.notnull().all()
    print_id(indivi)
    indivi['quifoy_bis'] = 2
    indivi.loc[indivi.quifoy == 'vous', 'quifoy_bis'] = 0
    indivi.loc[indivi.quifoy == 'conj', 'quifoy_bis'] = 1
    indivi.loc[indivi.quifoy == 'pac', 'quifoy_bis'] = 2
    del indivi['quifoy']
    indivi['quifoy'] = indivi.quifoy_bis.copy()
    del indivi['quifoy_bis']

    print_id(indivi)
    pac = indivi.loc[indivi['quifoy'] == 2, ['quifoy', 'idfoy', 'noindiv']].copy()
    print_id(pac)

    j = 2
    while pac.duplicated(['quifoy', 'idfoy']).any():
        pac.loc[pac.duplicated(['quifoy', 'idfoy']), 'quifoy'] = j
        j += 1

    print_id(pac)
    indivi = indivi.merge(pac, on = ['noindiv', 'idfoy'], how = "left")
    indivi['quifoy'] = indivi['quifoy_x']
    indivi['quifoy'] = where(indivi['quifoy_x'] == 2, indivi['quifoy_y'], indivi['quifoy_x'])
    del indivi['quifoy_x'], indivi['quifoy_y']
    print_id(indivi)
    del pac
    assert len(indivi[indivi.duplicated(subset = ['idfoy', 'quifoy'])]) == 0, \
        "Il y a {} doublons idfoy/quifoy".format(
            len(indivi[indivi.duplicated(subset = ['idfoy', 'quifoy'])])
            )
    print_id(indivi)

    log.info("Etape 8 : création des fichiers totaux")
    famille = temporary_store['famc_{}'.format(year)]

    log.info("    8.1 : création de tot2 & tot3")
    tot2 = indivi.merge(famille, on = 'noindiv', how = 'inner')
    # TODO: MBJ increase in number of menage/foyer when merging with family ...
    del famille

    control(tot2, debug = True, verbose = True)
    assert tot2.quifam.notnull().all()

    temporary_store['tot2_{}'.format(year)] = tot2
    del indivi
    log.info("    tot2 saved")
    tot2 = tot2[tot2.idmen.notnull()].copy()

    print_id(tot2)
    tot3 = tot2
    # TODO: check where they come from
    log.info("Avant élimination des doublons noindiv: {}".format(len(tot3)))

    tot3 = tot3.drop_duplicates(subset = 'noindiv')
    log.info("Après élimination des doublons noindiv: {}".format(len(tot3)))

    # Block to remove any unwanted duplicated pair

    control(tot3, debug = True, verbose = True)
    tot3 = tot3.drop_duplicates(subset = ['idfoy', 'quifoy'])
    log.info("Après élimination des doublons idfoy, quifoy: {}".format(len(tot3)))
    tot3 = tot3.drop_duplicates(subset = ['idfam', 'quifam'])
    log.info("Après élimination des doublons idfam, 'quifam: {}".format(len(tot3)))
    tot3 = tot3.drop_duplicates(subset = ['idmen', 'quimen'])
    log.info("Après élimination des doublons idmen, quimen: {}".format(len(tot3)))
    tot3 = tot3.drop_duplicates(subset = ['noindiv'])
    control(tot3)

    log.info("    8.2 : On ajoute les variables individualisables")

    allvars = temporary_store['ind_vars_to_remove_{}'.format(year)]
    vars2 = set(tot3.columns).difference(set(allvars))
    tot3 = tot3[list(vars2)]
    log.info("{}".format(len(tot3)))

    assert not(tot3.duplicated(subset = ['noindiv']).any()), "doublon dans tot3['noindiv']"
    lg_dup = len(tot3[tot3.duplicated(['idfoy', 'quifoy'])])
    assert lg_dup == 0, "{} pairs of idfoy/quifoy in tot3 are duplicated".format(lg_dup)

    temporary_store['tot3_{}'.format(year)] = tot3
    control(tot3)

    del tot2, allvars, tot3, vars2
    gc.collect()
    log.info("tot3 sauvegardé")


@temporary_store_decorator(file_name = 'erfs')
def create_final(temporary_store = None, year = None):

    assert temporary_store is not None
    assert year is not None

    log.info("création de final")
    foy_ind = temporary_store['foy_ind_{}'.format(year)]
    tot3 = temporary_store['tot3_{}'.format(year)]

    log.info("Stats on tot3")
    print_id(tot3)
    log.info("Stats on foy_ind")
    print_id(foy_ind)
    foy_ind.set_index(['idfoy', 'quifoy'], inplace = True, verify_integrity = True)
    tot3.set_index(['idfoy', 'quifoy'], inplace = True, verify_integrity = True)

    # tot3 = concat([tot3, foy_ind], join_axes=[tot3.index], axis=1, verify_integrity = True)

    # TODO improve this
    foy_ind.drop([u'alr', u'rsti', u'sali', u'choi'], axis = 1, inplace = True)
    tot3 = tot3.join(foy_ind)
    tot3.reset_index(inplace = True)
    foy_ind.reset_index(inplace = True)

    # tot3 = tot3.drop_duplicates(subset=['idfam', 'quifam'])
    control(tot3, verbose=True)
    final = tot3.loc[tot3.idmen.notnull(), :].copy()

    control(final, verbose=True)
    del tot3, foy_ind
    gc.collect()

    log.info("    loading fip")
    sif = temporary_store['sif_{}'.format(year)]
    log.info("Columns from sif dataframe: {}".format(sif.columns))
    log.info("    update final using fip")
    final.set_index('noindiv', inplace = True, verify_integrity = True)

    # TODO: IL FAUT UNE METHODE POUR GERER LES DOUBLES DECLARATIONS
    #  On ne garde que les sif.noindiv qui correspondent à des idfoy == "vous"
    #  Et on enlève les duplicates
    idfoys = final.loc[final.quifoy == 0, "idfoy"]
    sif = sif[sif.noindiv.isin(idfoys) & ~(sif.change.isin(['M', 'S', 'Z']))].copy()
    sif.drop_duplicates(subset = ['noindiv'], inplace = True)
    sif.set_index('noindiv', inplace = True, verify_integrity = True)
    final = final.join(sif)
    final.reset_index(inplace = True)

    control(final, debug=True)

    final['caseP'] = final.caseP.fillna(False)
    final['caseF'] = final.caseF.fillna(False)
    print_id(final)

    temporary_store['final_{}'.format(year)] = final
    log.info("final sauvegardé")
    del sif, final

if __name__ == '__main__':
    year = 2009
    logging.basicConfig(level = logging.INFO, filename = 'step_06.log', filemode = 'w')
    create_totals_first_pass(year = year)
    create_totals_second_pass(year = year)
    create_final(year = year)
    log.info("étape 06 remise en forme des données terminée")



from typing import Any, Optional

from multipledispatch import dispatch  # type: ignore

from openfisca_core.model_api import Variable, ADD, YEAR
from openfisca_core.reforms import Reform  # type: ignore
from openfisca_core.taxbenefitsystems import TaxBenefitSystem  # type: ignore

from openfisca_france.entities import Individu, FoyerFiscal, Menage
from openfisca_france_data.erfs_fpr.scenario import ErfsFprSurveyScenario
from openfisca_france_data import france_data_tax_benefit_system


from openfisca_france_data.model.id_variables import (
    idmen_original,
    noindiv,
    )

variables_converted_to_annual = [
    "salaire_imposable",
    "chomage_imposable",
    "retraite_imposable",
    "salaire_net",
    "chomage_net",
    "retraite_nette",
    "ppa",
    ]


menage_projected_variables = [

]


class erfs_fpr_plugin(Reform):
    name = "ERFS-FPR ids plugin"

    def apply(self):

        for variable in variables_converted_to_annual:
            class_name =  f"{variable}_annuel"
            label = f"{variable} sur l'année entière"

            def annual_formula_creator(variable):
                def formula(individu, period):
                    result = individu(variable, period, options = [ADD])
                    return result

                formula.__name__ = 'formula'

                return formula

            variable_instance = type(class_name, (Variable,), dict(
                value_type = float,
                entity = self.variables[variable].entity,
                label = label,
                definition_period = YEAR,
                formula = annual_formula_creator(variable),
                ))

            self.add_variable(variable_instance)
            del variable_instance

        for variable in menage_projected_variables:
            class_name =  f"{variable}_menage"
            label = f"{variable} agrégée à l'échelle du ménage"

            def projection_formula_creator(variable):
                def formula(menage, period):
                    result_i = menage.members.foyer_fiscal(variable, period, options = [ADD])
                    result = menage.sum(result_i, role = FoyerFiscal.DECLARANT_PRINCIPAL)
                    return result

                formula.__name__ = 'formula'

                return formula

            variable_instance = type(class_name, (Variable,), dict(
                value_type = float,
                entity = Menage,
                label = label,
                definition_period = YEAR,
                formula = projection_formula_creator(variable),
                ))

            self.add_variable(variable_instance)
            del variable_instance


        self.add_variable(idmen_original)
        self.add_variable(noindiv)


def get_survey_scenario(
        year: int = None,
        rebuild_input_data: bool = False,
        tax_benefit_system: Optional[TaxBenefitSystem] = None,
        baseline_tax_benefit_system: Optional[TaxBenefitSystem] = None,
        data: Any = None,  # Plusiers formats possibles.
        reform: Optional[Reform] = None,
        use_marginal_tax_rate: bool = False,
        variation_factor: float = 0.03,
        varying_variable: str = None,
        survey_name: str = "input",
        ) -> ErfsFprSurveyScenario:
    """Helper pour créer un `ErfsFprSurveyScenario`.

    :param year:                        L'année des données de l'enquête.
    :param rebuild_input_data:          Si l'on doit formatter les données (raw) ou pas.
    :param tax_benefit_system:          Le `TaxBenefitSystem` déjà réformé.
    :param baseline_tax_benefit_system: Le `TaxBenefitSystem` au droit courant.
    :param data:                        Les données de l'enquête.
    :param reform:                      Une réforme à appliquer à *france_data_tax_benefit_system*.
    """
    assert year is not None
    tax_benefit_system = get_tax_benefit_system(
        tax_benefit_system,
        reform,
        )

    baseline_tax_benefit_system = get_baseline_tax_benefit_system(
        baseline_tax_benefit_system,
        )

    if not use_marginal_tax_rate:
        survey_scenario = ErfsFprSurveyScenario.create(
            tax_benefit_system = tax_benefit_system,
            baseline_tax_benefit_system = baseline_tax_benefit_system,
            year = year,
            )
    else:
        assert varying_variable is not None, "You need to specify the varying variable."
        survey_scenario = ErfsFprSurveyScenario.create(
            tax_benefit_system = tax_benefit_system,
            baseline_tax_benefit_system = baseline_tax_benefit_system,
            year = year,
            )
        # taux marginaux !!
        survey_scenario.variation_factor = variation_factor
        survey_scenario.varying_variable = varying_variable

    # S'il n'y a pas de données, on sait où les trouver.
    if data is None:
        input_data_table_by_entity = dict(
            individu = f"individu_{year}",
            menage = f"menage_{year}",
            )

        input_data_table_by_entity_by_period = dict()
        input_data_table_by_entity_by_period[year] = input_data_table_by_entity

        data = dict(
            input_data_table_by_entity_by_period = input_data_table_by_entity_by_period,
            survey = survey_name
            )

    # Les données peuvent venir en différents formats :
    #
    #  * Pandas' DataFrame
    #  * CSV
    #  * HDF5
    #  * Etc.
    #
    #  .. seealso:: :func:`ErfsFprSurveyScenario.init_from_data`
    survey_scenario.init_from_data(
        data = data,
        rebuild_input_data = rebuild_input_data,
        use_marginal_tax_rate = use_marginal_tax_rate,
        )

    return survey_scenario


# Appelé quand *tax_benefit_system* est un :class:`TaxBenefitSystem`
@dispatch(TaxBenefitSystem, object)  # noqa: F811
def get_tax_benefit_system(
        tax_benefit_system: TaxBenefitSystem,
        reform: Optional[None],
        ) -> TaxBenefitSystem:
    return tax_benefit_system


# Appelé quand *tax_benefit_system* est `None`, mais *reform* est une :class:`Reform`
@dispatch(object, Reform)  # noqa: F811
def get_tax_benefit_system(
        tax_benefit_system: None,
        reform: Reform,
        ) -> TaxBenefitSystem:
    return reform(erfs_fpr_plugin(france_data_tax_benefit_system))


# Appelé quand *tax_benefit_system* et *reform* sont `None`
@dispatch(object, object)  # noqa: F811
def get_tax_benefit_system(
        tax_benefit_system: None,
        reform: None,
        ) -> TaxBenefitSystem:
    return erfs_fpr_plugin(france_data_tax_benefit_system)


# Appelé quand *tax_benefit_system* est un :class:`TaxBenefitSystem`
@dispatch(TaxBenefitSystem)  # noqa: F811
def get_baseline_tax_benefit_system(
        tax_benefit_system: TaxBenefitSystem,
        ) -> TaxBenefitSystem:
    return tax_benefit_system


# Appelé quand *tax_benefit_system* est `None`
@dispatch(object)  # noqa: F811
def get_baseline_tax_benefit_system(
        tax_benefit_system: None,
        ) -> TaxBenefitSystem:
    return erfs_fpr_plugin(france_data_tax_benefit_system)



import collections
import logging
import json
from pathlib import Path

import numpy as np
import pandas as pd
import pkg_resources


from openfisca_survey_manager.aggregates import AbstractAggregates
from openfisca_france_data import AGGREGATES_DEFAULT_VARS  # type: ignore


log = logging.getLogger(__name__)


class FranceAggregates(AbstractAggregates):
    labels = collections.OrderedDict((
        ('label', "Mesure"),
        ('entity', "Entité"),
        ('reform_amount', "Dépenses\n(millions d'€)"),
        ('reform_beneficiaries', "Bénéficiaires\n(milliers)"),
        ('baseline_amount', "Dépenses initiales\n(millions d'€)"),
        ('baseline_beneficiaries', "Bénéficiaires\ninitiaux\n(milliers)"),
        ('actual_amount', "Dépenses\nréelles\n(millions d'€)"),
        ('actual_beneficiaries', "Bénéficiaires\nréels\n(milliers)"),
        ('amount_absolute_difference', "Diff. absolue\nDépenses\n(millions d'€)"),
        ('beneficiaries_absolute_difference', "Diff absolue\nBénéficiaires\n(milliers)"),
        ('amount_relative_difference', "Diff. relative\nDépenses"),
        ('beneficiaries_relative_difference', "Diff. relative\nBénéficiaires"),
        ))
    aggregate_variables = AGGREGATES_DEFAULT_VARS
    target_source = None

    def __init__(self, survey_scenario = None, target_source = None):
        super().__init__(survey_scenario = survey_scenario)
        self.target_source = target_source

    def load_actual_data(self, year = None):
        target_source = self.target_source
        assert target_source in ["ines", "taxipp"], "les options possible pour source_cible sont ines ou taxipp"
        assert year is not None

        if target_source == "taxipp":
            taxipp_aggregates_file = Path(
                pkg_resources.get_distribution("openfisca-france_data").location,
                "openfisca_france_data",
                "assets",
                "aggregats",
                "taxipp",
                "agregats_tests_taxipp_2_0.xlsx"
                )
            df = (
                pd.read_excel(
                    taxipp_aggregates_file,
                    # "https://gitlab.com/ipp/partage-public-ipp/taxipp/-/blob/master/simulation/assets/agregats_tests_taxipp_2_0.xlsx",
                    # engine = 'openpyxl'
                    )
                .rename(columns = str.lower)
                .rename(columns = {"unnamed: 0": "description"})
                .dropna(subset = ["annee 2019", "annee 2018", "annee 2017", "annee 2016"], how = "all")
                )
            if f"annee {year}" not in df:
                return

            df = (
                df[["variable_openfisca", f"annee {year}"]]
                .dropna()
                .rename(columns = {
                    "variable_openfisca": "variable",
                    f"annee {year}": year,
                    })
                )

            beneficiaries = (
                df.loc[df.variable.str.startswith("nombre")]
                .set_index("variable")
                .rename(index = lambda x : x.replace("nombre_", ""))
                .rename(columns = {year: "actual_beneficiaries"})
                ) / self.beneficiaries_unit

            amounts = (
                df.loc[~df.variable.str.startswith("nombre")]
                .set_index("variable")
                .rename(columns = {year: "actual_amount"})
                ) / self.amount_unit

            result = amounts.merge(beneficiaries, on = "variable", how = "outer").drop("PAS SIMULE")

        elif target_source == "ines":
            ines_aggregates_file = Path(
                pkg_resources.get_distribution("openfisca-france_data").location,
                "openfisca_france_data",
                "assets",
                "aggregats",
                "ines",
                f"ines_{year}.json"
                )

            with open(ines_aggregates_file, 'r') as f:
                data = json.load(f)

            result = pd.DataFrame(data['data']).drop('notes', axis = 1)
            result['actual_beneficiaries'] = result. actual_beneficiaries / self.beneficiaries_unit
            result['actual_amount'] = result. actual_amount / self.amount_unit

            result = result[["variable","actual_amount","actual_beneficiaries"]].set_index("variable")

        return result



import pytest

from openfisca_core.reforms import Reform  # type: ignore
from openfisca_france import FranceTaxBenefitSystem as TaxBenefitSystem  # type: ignore

from openfisca_france.reforms.plf2015 import plf2015  # type: ignore

from openfisca_france_data import france_data_tax_benefit_system
from openfisca_france_data.erfs_fpr.get_survey_scenario import (
    get_survey_scenario,
    get_tax_benefit_system,
    get_baseline_tax_benefit_system,
    )


@pytest.fixture
def tax_benefit_system() -> TaxBenefitSystem:
    return TaxBenefitSystem()


@pytest.fixture
def default_tax_benefit_system() -> TaxBenefitSystem:
    return france_data_tax_benefit_system


@pytest.fixture
def reform() -> Reform:
    return plf2015


@pytest.mark.skip(reason = 'AssertionError: {} is not a valid path')
def test_get_survey_scenario():
    assert get_survey_scenario()


def test_get_tax_benefit_system(tax_benefit_system, reform):
    assert get_tax_benefit_system(tax_benefit_system, reform), tax_benefit_system


def test_get_tax_benefit_system_when_tax_benefit_system(tax_benefit_system):
    assert get_tax_benefit_system(tax_benefit_system, None), tax_benefit_system


def test_get_tax_benefit_system_when_reform(reform, default_tax_benefit_system):
    assert get_tax_benefit_system(None, reform), reform(default_tax_benefit_system)


def test_get_tax_benefit_system_when_none(default_tax_benefit_system):
    assert get_tax_benefit_system(None, None), default_tax_benefit_system


def test_get_baseline_tax_benefit_system(tax_benefit_system):
    assert get_baseline_tax_benefit_system(tax_benefit_system), tax_benefit_system


def test_get_baseline_tax_benefit_system_when_none(default_tax_benefit_system):
    assert get_baseline_tax_benefit_system(None), default_tax_benefit_system



from openfisca_france.model.base import *


class Deciles(Enum):
    __order__ = 'hors_champs decile_1 decile_2 decile_3 decile_4 decile_5 decile_6 decile_7 decile_8 decile_9 decile_10'  # Needed to keep the order in Python 2
    hors_champs = "Hors champ"
    decile_1 = "1er décile"
    decile_2 = "2nd décile"
    decile_3 = "3e décile"
    decile_4 = "4e décile"
    decile_5 = "5e décile"
    decile_6 = "6e décile"
    decile_7 = "7e décile"
    decile_8 = "8e décile"
    decile_9 = "9e décile"
    decile_10 = "10e décile"
